@@ -10,18 +10,21 @@
 // FIXME: Do not be lazy.
 #![allow(dead_code)]
 
-use std::os::fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd};
+use std::{
+    os::fd::{AsFd, AsRawFd, FromRawFd, OwnedFd, RawFd},
+    sync::LazyLock,
+};
 
 use libseccomp::ScmpSyscall;
 use nix::{
     errno::Errno,
     fcntl::{AtFlags, OpenHow},
+    sys::socket::{bind, connect, SockFlag, SockaddrLike},
     unistd::UnlinkatFlags,
     NixPath,
 };
-use once_cell::sync::Lazy;
 
-use crate::{fs::fillrandom, kernel::rename::RenameFlags, path::XPath};
+use crate::{compat::RenameFlags, fs::fillrandom, path::XPath};
 
 /// A platform-sized secure cookie
 ///
@@ -56,116 +59,151 @@ pub(crate) fn getcookie() -> Result<Cookie, Errno> {
 
 // These cookies are confined by seccomp for use with openat2(2).
 #[allow(clippy::disallowed_methods)]
-pub(crate) static OPENAT2_COOKIE_ARG4: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static OPENAT2_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static OPENAT2_COOKIE_ARG5: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static OPENAT2_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: socket(2) $arg3
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SOCKET_COOKIE_ARG3: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
+pub static SOCKET_COOKIE_ARG3: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: socket(2) $arg4
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SOCKET_COOKIE_ARG4: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
+pub static SOCKET_COOKIE_ARG4: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: socket(2) $arg5
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SOCKET_COOKIE_ARG5: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
+pub static SOCKET_COOKIE_ARG5: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: accept4(2) $arg4
 #[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG2: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static ACCEPT4_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: accept4(2) $arg5
 #[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static ACCEPT4_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: bind(2) $arg3
 #[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static BIND_COOKIE_ARG3: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: bind(2) $arg4
 #[allow(clippy::disallowed_methods)]
-pub(crate) static MEMFD_CREATE_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static BIND_COOKIE_ARG4: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: bind(2) $arg5
 #[allow(clippy::disallowed_methods)]
-pub(crate) static RENAMEAT2_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static BIND_COOKIE_ARG5: LazyLock<Cookie> = LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: connect(2) $arg3
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG2: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static CONNECT_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: connect(2) $arg4
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static CONNECT_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+/// System call argument cookie: connect(2) $arg5
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub static CONNECT_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static MEMFD_CREATE_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE64_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static MEMFD_CREATE_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE64_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static MEMFD_CREATE_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static TRUNCATE64_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static MEMFD_CREATE_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG2: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static RENAMEAT2_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE64_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE64_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE64_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static FTRUNCATE64_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static UNLINKAT_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static TRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static UNLINKAT_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static FTRUNCATE_COOKIE_ARG2: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static UNLINKAT_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static FTRUNCATE_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static LINKAT_COOKIE_ARG5: Lazy<Cookie> = Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static FTRUNCATE_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE64_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE64_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static FTRUNCATE64_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static UNLINKAT_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static UNLINKAT_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static UNLINKAT_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
+#[allow(clippy::disallowed_methods)]
+pub(crate) static LINKAT_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 
 /// These are used in `syd::fs::seccomp_notify_addfd`.
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 
 /// These are used in `syd::fs::seccomp_notify_respond`.
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 
 /// These are used in `syd::proc::procmap_query`.
 #[allow(clippy::disallowed_methods)]
-pub(crate) static PROCMAP_QUERY_COOKIE_ARG3: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static PROCMAP_QUERY_COOKIE_ARG3: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static PROCMAP_QUERY_COOKIE_ARG4: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static PROCMAP_QUERY_COOKIE_ARG4: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 #[allow(clippy::disallowed_methods)]
-pub(crate) static PROCMAP_QUERY_COOKIE_ARG5: Lazy<Cookie> =
-    Lazy::new(|| getcookie().expect("getcookie"));
+pub(crate) static PROCMAP_QUERY_COOKIE_ARG5: LazyLock<Cookie> =
+    LazyLock::new(|| getcookie().expect("getcookie"));
 
 /// Safe openat2 confined by syscall cookies.
 #[inline(always)]
@@ -197,7 +235,7 @@ pub(crate) fn safe_openat2<Fd: AsFd>(
 }
 
 /// socket(2) may be multiplexed by socketcall(2).
-pub(crate) static SYS_SOCKET: Lazy<Option<libc::c_long>> = Lazy::new(|| {
+pub(crate) static SYS_SOCKET: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("socket")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -209,9 +247,9 @@ pub(crate) static SYS_SOCKET: Lazy<Option<libc::c_long>> = Lazy::new(|| {
     }
 });
 
-/// Safe socket confined by syscall cookies.
+/// Safe socket(2) confined by syscall cookies.
 #[inline(always)]
-pub(crate) fn safe_socket(
+pub fn safe_socket(
     domain: libc::c_int,
     stype: libc::c_int,
     proto: libc::c_int,
@@ -243,6 +281,133 @@ pub(crate) fn safe_socket(
     })
 }
 
+/// accept4(2) may be multiplexed by socketcall(2).
+pub(crate) static SYS_ACCEPT4: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+    match ScmpSyscall::from_name("accept4")
+        .map(i32::from)
+        .map(libc::c_long::from)
+        .ok()
+    {
+        Some(n) if n < 0 => None,
+        Some(n) => Some(n),
+        None => None,
+    }
+});
+
+/// Safe accept4(2) confined by syscall cookies.
+///
+/// # Safety
+///
+/// Dereferences raw pointers `addr` and `len`.
+#[inline(always)]
+pub unsafe fn safe_accept4<Fd: AsFd>(
+    fd: Fd,
+    addr: *mut libc::sockaddr,
+    len: *mut libc::socklen_t,
+    flags: SockFlag,
+) -> Result<OwnedFd, Errno> {
+    if let Some(sys_accept4) = *SYS_ACCEPT4 {
+        // SAFETY: In libc we trust.
+        #[allow(clippy::cast_possible_truncation)]
+        Errno::result(unsafe {
+            libc::syscall(
+                sys_accept4,
+                fd.as_fd().as_raw_fd(),
+                addr,
+                len,
+                flags.bits(),
+                *ACCEPT4_COOKIE_ARG4,
+                *ACCEPT4_COOKIE_ARG5,
+            )
+        })
+        .map(|fd| {
+            // SAFETY: accept4 returns a valid fd on success.
+            unsafe { OwnedFd::from_raw_fd(fd as RawFd) }
+        })
+    } else {
+        // SAFETY: socketcall(2) on multiplexed architecture.
+        Errno::result(unsafe { libc::accept4(fd.as_fd().as_raw_fd(), addr, len, flags.bits()) })
+            .map(|fd| {
+                // SAFETY: accept4 returns a valid fd on success.
+                unsafe { OwnedFd::from_raw_fd(fd) }
+            })
+    }
+}
+
+/// bind(2) may be multiplexed by socketcall(2).
+pub(crate) static SYS_BIND: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+    match ScmpSyscall::from_name("bind")
+        .map(i32::from)
+        .map(libc::c_long::from)
+        .ok()
+    {
+        Some(n) if n < 0 => None,
+        Some(n) => Some(n),
+        None => None,
+    }
+});
+
+/// Safe bind(2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_bind<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno> {
+    if let Some(sys_bind) = *SYS_BIND {
+        // SAFETY: In libc we trust.
+        #[allow(clippy::cast_possible_truncation)]
+        Errno::result(unsafe {
+            libc::syscall(
+                sys_bind,
+                fd.as_fd().as_raw_fd(),
+                addr.as_ptr(),
+                addr.len(),
+                *BIND_COOKIE_ARG3,
+                *BIND_COOKIE_ARG4,
+                *BIND_COOKIE_ARG5,
+            )
+        })
+        .map(drop)
+    } else {
+        // socketcall(2) on multiplexed architecture.
+        bind(fd.as_fd().as_raw_fd(), addr)
+    }
+}
+
+/// connect(2) may be multiplexed by socketcall(2).
+pub(crate) static SYS_CONNECT: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
+    match ScmpSyscall::from_name("connect")
+        .map(i32::from)
+        .map(libc::c_long::from)
+        .ok()
+    {
+        Some(n) if n < 0 => None,
+        Some(n) => Some(n),
+        None => None,
+    }
+});
+
+/// Safe connect(2) confined by syscall cookies.
+#[inline(always)]
+pub fn safe_connect<Fd: AsFd>(fd: Fd, addr: &dyn SockaddrLike) -> Result<(), Errno> {
+    if let Some(sys_connect) = *SYS_CONNECT {
+        // SAFETY: In libc we trust.
+        #[allow(clippy::cast_possible_truncation)]
+        Errno::result(unsafe {
+            libc::syscall(
+                sys_connect,
+                fd.as_fd().as_raw_fd(),
+                addr.as_ptr(),
+                addr.len(),
+                *CONNECT_COOKIE_ARG3,
+                *CONNECT_COOKIE_ARG4,
+                *CONNECT_COOKIE_ARG5,
+            )
+        })
+        .map(drop)
+    } else {
+        // socketcall(2) on multiplexed architecture.
+        connect(fd.as_fd().as_raw_fd(), addr)
+    }
+}
+
 /// Safe memfd_create confined by syscall cookies.
 #[inline(always)]
 pub(crate) fn safe_memfd_create(name: &[u8], flags: libc::c_uint) -> Result<OwnedFd, Errno> {
@@ -294,20 +459,21 @@ pub(crate) fn safe_renameat2<Fd1: AsFd, Fd2: AsFd>(
 }
 
 /// truncate(2) may be aliased to truncate64(2) by libc.
-static SYS_TRUNCATE: Lazy<Option<libc::c_long>> = Lazy::new(|| {
-    match ScmpSyscall::from_name("truncate")
-        .map(i32::from)
-        .map(libc::c_long::from)
-        .ok()
-    {
-        Some(n) if n < 0 => None,
-        Some(n) => Some(n),
-        None => None,
-    }
-});
+static SYS_TRUNCATE: LazyLock<Option<libc::c_long>> =
+    LazyLock::new(|| {
+        match ScmpSyscall::from_name("truncate")
+            .map(i32::from)
+            .map(libc::c_long::from)
+            .ok()
+        {
+            Some(n) if n < 0 => None,
+            Some(n) => Some(n),
+            None => None,
+        }
+    });
 
 /// truncate64(2) may not always be available via libc.
-static SYS_TRUNCATE64: Lazy<Option<libc::c_long>> = Lazy::new(|| {
+static SYS_TRUNCATE64: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("truncate64")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -320,7 +486,7 @@ static SYS_TRUNCATE64: Lazy<Option<libc::c_long>> = Lazy::new(|| {
 });
 
 /// ftruncate(2) may be aliased to ftruncate64(2) by libc.
-static SYS_FTRUNCATE: Lazy<Option<libc::c_long>> = Lazy::new(|| {
+static SYS_FTRUNCATE: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("ftruncate")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -333,7 +499,7 @@ static SYS_FTRUNCATE: Lazy<Option<libc::c_long>> = Lazy::new(|| {
 });
 
 /// ftruncate64(2) may not always be available via libc.
-static SYS_FTRUNCATE64: Lazy<Option<libc::c_long>> = Lazy::new(|| {
+static SYS_FTRUNCATE64: LazyLock<Option<libc::c_long>> = LazyLock::new(|| {
     match ScmpSyscall::from_name("ftruncate64")
         .map(i32::from)
         .map(libc::c_long::from)
@@ -369,9 +535,11 @@ pub(crate) fn safe_truncate(path: &XPath, len: libc::off_t) -> Result<(), Errno>
 pub(crate) fn safe_truncate64(path: &XPath, len: libc::off64_t) -> Result<(), Errno> {
     #[cfg(not(any(
         target_pointer_width = "64",
+        all(target_arch = "x86_64", target_pointer_width = "32"),
         target_arch = "x86",
         target_arch = "arm",
         target_arch = "powerpc",
+        target_arch = "m68k",
         target_arch = "mips",
         target_arch = "mips32r6",
     )))]
@@ -379,12 +547,15 @@ pub(crate) fn safe_truncate64(path: &XPath, len: libc::off64_t) -> Result<(), Er
         compile_error!("BUG: safe_truncate64 is not implemented for this architecture!");
     }
 
-    #[cfg(target_pointer_width = "64")]
+    #[cfg(any(
+        target_pointer_width = "64",
+        all(target_arch = "x86_64", target_pointer_width = "32"),
+    ))]
     {
         safe_truncate(path, len)
     }
 
-    #[cfg(target_arch = "x86")]
+    #[cfg(any(target_arch = "m68k", target_arch = "x86",))]
     {
         let sys_truncate64 = SYS_TRUNCATE64.ok_or(Errno::ENOSYS)?;
 
@@ -461,9 +632,11 @@ pub(crate) fn safe_ftruncate<Fd: AsFd>(fd: Fd, len: libc::off_t) -> Result<(), E
 pub(crate) fn safe_ftruncate64<Fd: AsFd>(fd: Fd, len: libc::off64_t) -> Result<(), Errno> {
     #[cfg(not(any(
         target_pointer_width = "64",
+        all(target_arch = "x86_64", target_pointer_width = "32"),
         target_arch = "x86",
         target_arch = "arm",
         target_arch = "powerpc",
+        target_arch = "m68k",
         target_arch = "mips",
         target_arch = "mips32r6",
     )))]
@@ -471,12 +644,15 @@ pub(crate) fn safe_ftruncate64<Fd: AsFd>(fd: Fd, len: libc::off64_t) -> Result<(
         compile_error!("BUG: safe_ftruncate64 is not implemented for this architecture!");
     }
 
-    #[cfg(target_pointer_width = "64")]
+    #[cfg(any(
+        target_pointer_width = "64",
+        all(target_arch = "x86_64", target_pointer_width = "32"),
+    ))]
     {
         safe_ftruncate(fd, len)
     }
 
-    #[cfg(target_arch = "x86")]
+    #[cfg(any(target_arch = "m68k", target_arch = "x86",))]
     {
         let sys_ftruncate64 = SYS_FTRUNCATE64.ok_or(Errno::ENOSYS)?;
 
