@@ -6,19 +6,22 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
+use std::os::fd::AsRawFd;
+
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
+    fcntl::{OFlag, ResolveFlag},
     sys::signal::{kill, Signal},
     unistd::Pid,
 };
 
 use crate::{
-    compat::{fstatx, statx},
+    compat::fstatx,
     config::PROC_FILE,
     debug, error,
-    fs::{readlinkat, CanonicalPath, FileInfo, FsFlags},
-    hook::{RemoteProcess, SysArg, SysFlags, UNotifyEventRequest},
+    fs::{readlinkat, safe_open_path_msym, CanonicalPath, FileInfo, FsFlags},
+    hook::{RemoteProcess, SysArg, UNotifyEventRequest},
     kernel::{sandbox_path, syscall_path_handler},
     path::XPathBuf,
     ptrace::{ptrace_get_error, ptrace_syscall_info, ptrace_syscall_info_seccomp},
@@ -38,7 +41,7 @@ pub(crate) fn sysenter_chdir<'a>(
     };
 
     // SAFETY: Apply deny_dotdot as necessary for chdir.
-    if sandbox.deny_dotdot() {
+    if sandbox.flags.deny_dotdot() {
         arg.fsflags.insert(FsFlags::NO_RESOLVE_DOTDOT);
     }
 
@@ -104,6 +107,20 @@ pub(crate) fn sysexit_chdir(
     // SAFETY: Validate /proc/$pid/cwd against TOCTTOU!
     let mut pfd = XPathBuf::from_pid(pid);
     pfd.push(b"cwd");
+    let cwd = match safe_open_path_msym(PROC_FILE(), &pfd, OFlag::O_DIRECTORY, ResolveFlag::empty())
+    {
+        Ok(cwd) => cwd,
+        Err(errno) => {
+            // SAFETY: Failed to open FD,
+            // assume TOCTTOU: terminate the process.
+            error!("ctx": "chdir", "op": "open_dir_fd",
+                "msg": format!("failed to open dir-fd for `{path}': {errno}"),
+                "err": errno as i32, "pid": pid.as_raw(), "path": &path,
+                "tip": "check with SYD_LOG=debug and/or submit a bug report");
+            let _ = kill(pid, Some(Signal::SIGKILL));
+            return Err(Errno::ESRCH);
+        }
+    };
 
     #[allow(clippy::disallowed_methods)]
     let fd = path.dir.as_ref().unwrap();
@@ -114,25 +131,23 @@ pub(crate) fn sysexit_chdir(
             // SAFETY: Failed to stat FD,
             // assume TOCTTOU: terminate the process.
             error!("ctx": "chdir", "op": "fstat_dir_fd",
-                "err": format!("failed to fstat dir-fd for `{path}': {errno}"),
-                "pid": pid.as_raw(),
-                "path": &path,
-                "errno": errno as i32);
+                "msg": format!("failed to fstat dir-fd for `{path}': {errno}"),
+                "err": errno as i32, "pid": pid.as_raw(), "path": &path,
+                "tip": "check with SYD_LOG=debug and/or submit a bug report");
             let _ = kill(pid, Some(Signal::SIGKILL));
             return Err(Errno::ESRCH);
         }
     };
 
-    let stx_cwd = match statx(PROC_FILE(), &pfd, 0, FileInfo::mask()) {
+    let stx_cwd = match fstatx(&cwd, FileInfo::mask()) {
         Ok(stx) => stx,
         Err(errno) => {
             // SAFETY: Failed to stat CWD,
             // assume TOCTTOU: terminate the process.
             error!("ctx": "chdir", "op": "stat_cwd_symlink",
-                "err": format!("failed to stat cwd-symlink for `{path}': {errno}"),
-                "pid": pid.as_raw(),
-                "path": &path,
-                "errno": errno as i32);
+                "msg": format!("failed to stat cwd-symlink for `{path}': {errno}"),
+                "err": errno as i32, "pid": pid.as_raw(), "path": &path,
+                "tip": "check with SYD_LOG=debug and/or submit a bug report");
             let _ = kill(pid, Some(Signal::SIGKILL));
             return Err(Errno::ESRCH);
         }
@@ -154,14 +169,13 @@ pub(crate) fn sysexit_chdir(
     if !is_match {
         // SAFETY: CWD changed, which indicates
         // successful TOCTTOU attempt: terminate the process.
+        let pfd = XPathBuf::from_self_fd(cwd.as_raw_fd());
         let cwd = readlinkat(PROC_FILE(), &pfd)
             .ok()
             .unwrap_or_else(|| XPathBuf::from("?"));
         error!("ctx": "chdir", "op": "dir_mismatch",
-            "err": format!("dir mismatch detected for directory `{path}' -> `{cwd}': assume TOCTTOU!"),
-            "pid": pid.as_raw(),
-            "path": &path,
-            "real": cwd,
+            "msg": format!("dir mismatch detected for directory `{path}' -> `{cwd}': assume TOCTTOU!"),
+            "pid": pid.as_raw(), "path": &path, "real": cwd,
             "cwd_mount_id": stx_cwd.stx_mnt_id,
             "dir_mount_id": stx_fd.stx_mnt_id,
             "cwd_inode": stx_cwd.stx_ino,
@@ -183,9 +197,9 @@ pub(crate) fn sysexit_chdir(
     Ok(())
 }
 pub(crate) fn sys_chdir(request: UNotifyEventRequest) -> ScmpNotifResp {
+    // SAFETY: We do not pass `fsflags` which defaults to MUST_PATH.
     let argv = &[SysArg {
         path: Some(0),
-        flags: SysFlags::UNSAFE_CONT,
         ..Default::default()
     }];
 
@@ -210,12 +224,10 @@ pub(crate) fn sys_chdir(request: UNotifyEventRequest) -> ScmpNotifResp {
 }
 
 pub(crate) fn sys_fchdir(request: UNotifyEventRequest) -> ScmpNotifResp {
-    // SAFETY:
-    // 1. fchdir is fd-only, so UNSAFE_CONT is ok.
-    // 2. We do not pass `fsflags` which defaults to MUST_PATH.
+    // SAFETY: We do not pass `fsflags` which defaults to MUST_PATH.
+    // Note: fchdir works with O_PATH fds.
     let argv = &[SysArg {
         dirfd: Some(0),
-        flags: SysFlags::UNSAFE_CONT,
         ..Default::default()
     }];
 
