@@ -132,7 +132,7 @@ use libseccomp::ScmpNotifResp;
 use memchr::memmem;
 use nix::{
     errno::Errno,
-    fcntl::AtFlags,
+    fcntl::{AtFlags, OFlag},
     sys::{
         signal::{kill, Signal},
         stat::{Mode, SFlag},
@@ -141,7 +141,7 @@ use nix::{
 };
 
 use crate::{
-    fs::{to_valid_fd, CanonicalPath, FileInfo, FileType},
+    fs::{fd_status_flags, to_valid_fd, CanonicalPath, FileInfo, FileType},
     hook::{PathArgs, SysArg, UNotifyEventRequest},
     ioctl::{ioctl_names_get, Ioctl},
     notice,
@@ -198,10 +198,15 @@ pub(crate) fn sandbox_path(
 
     let mut action = Action::Allow;
     let mut filter = false;
-    let mut deny_errno = Errno::EACCES;
+    let check_caps = caps & Capability::CAP_PATH;
+    let mut deny_errno = if check_caps == Capability::CAP_WALK {
+        Errno::ENOENT
+    } else {
+        Errno::EACCES
+    };
 
     // Sandboxing.
-    for cap in caps & Capability::CAP_PATH {
+    for cap in check_caps {
         let (new_action, new_filter) = sandbox.check_path(cap, &path);
 
         if new_action >= action {
@@ -237,14 +242,17 @@ pub(crate) fn sandbox_path(
 
     if !filter && action >= Action::Warn {
         // Log warn for normal cases.
-        // Log info for path hiding unless explicitly specified to warn.
-        let is_warn = if caps != Capability::CAP_STAT {
-            true
-        } else {
-            !matches!(
+        // Log info for path hiding/walking unless explicitly specified to warn.
+        let is_warn = match caps {
+            Capability::CAP_STAT => !matches!(
                 sandbox.default_action(Capability::CAP_STAT),
                 Action::Filter | Action::Deny
-            )
+            ),
+            Capability::CAP_WALK => !matches!(
+                sandbox.default_action(Capability::CAP_WALK),
+                Action::Filter | Action::Deny
+            ),
+            _ => true,
         };
 
         if let Some(request) = request {
@@ -436,6 +444,11 @@ where
                     // as it may change after which is a TOCTOU vector.
                     let fd = request.get_fd(dirfd)?;
 
+                    // Validate WANT_READ against O_PATH.
+                    if arg.fsflags.want_read() && fd_status_flags(&fd)?.contains(OFlag::O_PATH) {
+                        return Err(Errno::EBADF);
+                    }
+
                     // Handle ftruncate etc. for files with encryption in progress.
                     let crypt_path = if crypt {
                         if let Ok(info) = FileInfo::from_fd(&fd) {
