@@ -8,8 +8,9 @@
 
 use std::{
     borrow::Cow,
+    env,
     fs::File,
-    io::{Seek, Write},
+    io::{BufWriter, Seek, Write},
     os::fd::{AsFd, AsRawFd, IntoRawFd, OwnedFd, RawFd},
 };
 
@@ -24,11 +25,17 @@ use nix::{
 
 use crate::{
     compat::{fstatx, STATX_MODE, STATX_UID},
-    config::{ESYD_SH, MAGIC_PREFIX, NULL_FD, PROC_FILE, ROOT_FILE, SAFE_MFD_FLAGS, SYD_EL},
+    config::{
+        ENV_ID, ESYD_SH, MAGIC_PREFIX, NULL_FD, PINK_FLOYD, PROC_FILE, ROOT_FILE, SAFE_MFD_FLAGS,
+        SEE_EMILY_PLAY, SYD_EL, THE_PIPER,
+    },
+    confine::EIDRM,
     cookie::safe_openat2,
+    err::err2no,
+    error,
     fs::{
-        create_memfd, is_dev_kfd, lock_fd, safe_open_how_magicsym, safe_open_magicsym, seal_memfd,
-        unlock_fd, CanonicalPath, FileInfo, FileType, FsFlags, MaybeFd,
+        create_memfd, format_oflags, is_dev_kfd, lock_fd, retry_on_eintr, safe_open_how_msym,
+        safe_open_msym, seal_memfd, unlock_fd, CanonicalPath, FileInfo, FileType, FsFlags, MaybeFd,
     },
     hash::aes_ctr_tmp,
     hook::{SysArg, UNotifyEventRequest},
@@ -89,8 +96,10 @@ pub(crate) fn sys_open(request: UNotifyEventRequest) -> ScmpNotifResp {
         Err(errno) => return request.fail_syscall(errno),
     };
 
-    // SAFETY: Reject undefined/invalid mode.
-    let mode = match to_mode(req.data.args[2], flags) {
+    // SAFETY:
+    // 1. Reject undefined/invalid mode.
+    // 2. Ignore mode if no creation flags are passed.
+    let mode = match to_mode(req.data.args[2], flags, true) {
         Ok(mode) => mode,
         Err(errno) => return request.fail_syscall(errno),
     };
@@ -110,8 +119,10 @@ pub(crate) fn sys_openat(request: UNotifyEventRequest) -> ScmpNotifResp {
         Err(errno) => return request.fail_syscall(errno),
     };
 
-    // SAFETY: Reject undefined/invalid mode.
-    let mode = match to_mode(req.data.args[3], flags) {
+    // SAFETY:
+    // 1. Reject undefined/invalid mode.
+    // 2. Ignore mode if no creation flags are passed.
+    let mode = match to_mode(req.data.args[3], flags, true) {
         Ok(mode) => mode,
         Err(errno) => return request.fail_syscall(errno),
     };
@@ -155,7 +166,10 @@ pub(crate) fn sys_openat2(request: UNotifyEventRequest) -> ScmpNotifResp {
         None => return request.fail_syscall(Errno::EINVAL),
     };
 
-    let mode = match to_mode(open_how.mode, flags) {
+    // SAFETY:
+    // 1. Reject undefined/invalid mode.
+    // 2. Reject non-zero mode if no creation flags are passed.
+    let mode = match to_mode(open_how.mode, flags, false) {
         Ok(mode) => mode,
         Err(errno) => return request.fail_syscall(errno),
     };
@@ -226,7 +240,13 @@ fn syscall_open_handler(
             return Err(Errno::ENOENT);
         }
 
-        let is_lock = sandbox.locked_for(req.pid());
+        // Check sandbox lock for Read access.
+        //
+        // a. LockState::Set allows no access.
+        // b. LockState::Read allows open(2) based read-only access.
+        //
+        // `is_lock` is true for LockState::Set and false for LockState::Read.
+        let is_lock = sandbox.locked_read_for(req.pid());
 
         // We use exactly one of Mktemp, Create, Truncate sandboxing
         // categories to improve usability, ie the priority is
@@ -302,6 +322,12 @@ fn syscall_open_handler(
                     Cow::Borrowed(SYD_EL)
                 } else if fack && fext.map(|ext| ext.is_equal(b"sh")).unwrap_or(false) {
                     Cow::Borrowed(ESYD_SH)
+                } else if path.abs().is_equal(b"/dev/syd/pink") {
+                    Cow::Borrowed(PINK_FLOYD)
+                } else if path.abs().is_equal(b"/dev/syd/pipe") {
+                    Cow::Borrowed(THE_PIPER)
+                } else if path.abs().is_equal(b"/dev/syd/play") {
+                    Cow::Borrowed(SEE_EMILY_PLAY)
                 } else {
                     return Err(Errno::EINVAL);
                 };
@@ -566,16 +592,16 @@ fn syscall_open_handler(
                     info!("ctx": "open", "op": "set_append_only",
                             "msg": "added O_APPEND and removed O_TRUNC from open flags",
                             "sys": request.syscall, "path": &path_check,
-                            "flags": format!("{flags:?}"),
-                            "cap": caps,
-                            "req": &request);
+                            "mode": mode.bits(),
+                            "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                            "type": path.typ, "cap": caps, "req": &request);
                 } else {
                     info!("ctx": "open", "op": "set_append_only",
                             "msg": "added O_APPEND and removed O_TRUNC from open flags",
                             "sys": request.syscall, "path": &path_check,
-                            "flags": format!("{flags:?}"),
-                            "cap": caps,
-                            "pid": request.scmpreq.pid);
+                            "mode": mode.bits(),
+                            "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                            "type": path.typ, "cap": caps, "pid": request.scmpreq.pid);
                 }
             }
 
@@ -584,11 +610,13 @@ fn syscall_open_handler(
 
         let force_umask = sandbox.umask;
         let verbose = sandbox.verbose;
-        let unsafe_kfd = sandbox.flags.allow_unsafe_kfd();
+        let unsafe_open_kfd = sandbox.flags.allow_unsafe_open_kfd();
         let unsafe_open_path = sandbox.flags.allow_unsafe_open_path();
         let restrict_create = !sandbox.flags.allow_unsafe_create();
+        let filter_machine_id =
+            !sandbox.flags.allow_unsafe_machine_id() && path.abs().is_machine_id();
         let filter_proc_status =
-            !sandbox.flags.allow_unsafe_proc_status() && path.abs().is_proc_pid_status();
+            !sandbox.flags.allow_unsafe_proc_pid_status() && path.abs().is_proc_pid_status();
         if sandbox.flags.force_cloexec() {
             o_cloexec = true;
         }
@@ -603,15 +631,17 @@ fn syscall_open_handler(
             if verbose {
                 warn!("ctx": "access", "cap": caps, "act": action,
                     "sys": request.syscall, "path": &path_check,
-                    "open_flags": format!("{flags:?}"),
-                    "open_mode": format!("{mode:?}"),
+                    "mode": mode.bits(),
+                    "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                    "type": path.typ,
                     "tip": format!("configure `allow/{grp}+{path_check}'"),
                     "req": &request);
             } else {
                 warn!("ctx": "access", "cap": caps, "act": action,
                     "sys": request.syscall, "path": &path_check,
-                    "open_flags": format!("{flags:?}"),
-                    "open_mode": format!("{mode:?}"),
+                    "mode": mode.bits(),
+                    "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                    "type": path.typ,
                     "tip": format!("configure `allow/{grp}+{path_check}'"),
                     "pid": request.scmpreq.pid);
             }
@@ -693,18 +723,34 @@ fn syscall_open_handler(
             // Return EACCES if:
             // 1. File is group or world-writable.
             // 2. File is not owned by caller.
+            let mut err = None;
             if mode & libc::S_IWOTH != 0 {
                 // File is world-writable.
-                return Err(Errno::EACCES);
-            }
-
-            if mode & libc::S_IWGRP != 0 {
+                err = Some("file is world writable");
+            } else if mode & libc::S_IWGRP != 0 {
                 // File is group-writable.
-                return Err(Errno::EACCES);
+                err = Some("file is group writable");
+            } else if statx.stx_uid != Uid::effective().as_raw() {
+                // File is not owned by caller.
+                err = Some("file is not owned by caller");
             }
 
-            if statx.stx_uid != Uid::effective().as_raw() {
-                // File is not owned by caller.
+            if let Some(msg) = err {
+                if verbose {
+                    error!("ctx": "trusted_file_creation", "cap": caps,
+                        "sys": request.syscall, "path": &path_check,
+                        "mode": mode, "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                        "msg": format!("create for untrusted path blocked: {msg}"),
+                        "tip": "fix path permissions or use `trace/allow_unsafe_create:1'",
+                        "req": &request);
+                } else {
+                    error!("ctx": "trusted_file_creation", "cap": caps,
+                        "sys": request.syscall, "path": &path_check,
+                        "mode": mode, "oflags": format_oflags(flags), "rflags": arg.fsflags,
+                        "msg": format!("create for untrusted path blocked: {msg}"),
+                        "tip": "fix path permissions or use `trace/allow_unsafe_create:1'",
+                        "pid": request.scmpreq.pid);
+                }
                 return Err(Errno::EACCES);
             }
         }
@@ -729,7 +775,7 @@ fn syscall_open_handler(
                     // SAFETY: Turn O_PATH flag to O_RDONLY for successful emulation.
                     flags.remove(OFlag::O_PATH);
                     flags.insert(OFlag::O_RDONLY);
-                } else if unsafe_kfd
+                } else if unsafe_open_kfd
                     && o_rdwr
                     && !o_creat
                     && !o_trunc
@@ -817,12 +863,13 @@ fn syscall_open_handler(
                         fd
                     } else {
                         // read-write encryption.
+                        //
                         // We do not need to send a response,
-                        // return a dummy response which will be
-                        // skipped by the handler.
-                        return Ok(ScmpNotifResp::new(0, 0, 0, 0));
+                        // send a dummy response to the caller
+                        // can skip it gracefully.
+                        return Ok(ScmpNotifResp::new(0, 0, EIDRM, 0));
                     }
-                } else if path.base.is_empty() {
+                } else if path.base.is_empty() && !o_tmpfl {
                     // Existing path: Construct path to /proc magic symlink,
                     // or pre-open file descriptor.
                     //
@@ -841,7 +888,7 @@ fn syscall_open_handler(
                     // Reopen the `O_PATH` path fd with the requested flags.
                     flags.remove(OFlag::O_CREAT);
                     flags.remove(OFlag::O_NOFOLLOW);
-                    let how = safe_open_how_magicsym(flags);
+                    let how = safe_open_how_msym(flags, ResolveFlag::empty());
 
                     // SAFETY: Record blocking call so it can get invalidated.
                     if is_blocking {
@@ -859,12 +906,40 @@ fn syscall_open_handler(
 
                     let mut fd = result?;
 
-                    // SAFETY: Filter security-sensitive proc_pid_status(5) contents.
-                    if filter_proc_status {
+                    if filter_machine_id {
+                        // SAFETY: Substitute machine-id(5) with per-Syd random boot-id.
+                        //
+                        // Generate machine-id(5) contents.
+                        #[allow(clippy::disallowed_methods)]
+                        let mut machine_id =
+                            env::var(ENV_ID).expect("BUG: SYD_ID environment variable isn't set!");
+                        machine_id.truncate(32);
+                        machine_id.push('\n');
+
                         // Open a memory fd to be sealed for write.
                         // Memory fd is exec-sealed on Linux>=6.3.
-                        let mut mfd =
-                            create_memfd(b"syd-pps\0", *SAFE_MFD_FLAGS).map(File::from)?;
+                        let mut mfd = create_memfd(b"syd-machine-id\0", *SAFE_MFD_FLAGS)
+                            .map(File::from)
+                            .map(BufWriter::new)?;
+
+                        // Write file contents, retrying on interrupts.
+                        mfd.write_all(machine_id.as_bytes())
+                            .map_err(|err| err2no(&err))?;
+                        // into_inner fails on flush errors.
+                        let mut mfd = mfd.into_inner().or(Err(Errno::EIO))?;
+
+                        // Rewind file offset and seal to deny further writes.
+                        mfd.rewind().or(Err(Errno::EIO))?;
+                        seal_memfd(&mfd)?;
+
+                        fd = mfd.into();
+                    } else if filter_proc_status {
+                        // SAFETY: Filter security-sensitive proc_pid_status(5) contents.
+                        //
+                        // Open a memory fd to be sealed for write.
+                        // Memory fd is exec-sealed on Linux>=6.3.
+                        let mut mfd = create_memfd(b"syd-proc-pid-status\0", *SAFE_MFD_FLAGS)
+                            .map(File::from)?;
 
                         // Filter proc_pid_status(5) into the memory fd.
                         mask_proc_pid_status(fd, &mfd)?;
@@ -916,7 +991,7 @@ fn do_open(path: &CanonicalPath, mut safe_flags: OFlag) -> Result<OwnedFd, Errno
         // Reopen the `O_PATH` path fd with the requested flags.
         safe_flags.remove(OFlag::O_NOFOLLOW);
 
-        safe_open_magicsym(PROC_FILE(), &pfd, safe_flags)
+        safe_open_msym(PROC_FILE(), &pfd, safe_flags, ResolveFlag::empty())
     } else {
         // Return a read-only fd to the underlying encrypted file.
         let how = OpenHow::new().flags(safe_flags).resolve(
@@ -979,8 +1054,8 @@ fn handle_crypt(
             flags.remove(OFlag::O_NOFOLLOW);
 
             let pfd = XPathBuf::from_self_fd(entry.crypt_fd);
-            match safe_open_magicsym(PROC_FILE(), &pfd, flags) {
-                Ok(fd) if lock_fd(&fd, false, false).is_ok() => {
+            match safe_open_msym(PROC_FILE(), &pfd, flags, ResolveFlag::empty()) {
+                Ok(fd) if retry_on_eintr(|| lock_fd(&fd, false, false)).is_ok() => {
                     // Return shared fd.
                     return Ok(Some(fd));
                 }
@@ -1034,7 +1109,7 @@ fn handle_crypt(
     // Similarly, `aes_ctr_tmp` respects non-blocking opens too,
     // and returns a non-blocking fd as necessary.
     let wait = !(safe_flags.contains(OFlag::O_NONBLOCK) || safe_flags.contains(OFlag::O_NDELAY));
-    lock_fd(&enc_fd, true, create || wait)?;
+    retry_on_eintr(|| lock_fd(&enc_fd, true, create || wait))?;
 
     // Acquire encryption sockets.
     let setup_fds = setup_fds?;
@@ -1066,7 +1141,7 @@ fn handle_crypt(
     );
     flags.insert(OFlag::O_RDWR);
 
-    let aes_fd = safe_open_magicsym(PROC_FILE(), &pfd, flags)?;
+    let aes_fd = safe_open_msym(PROC_FILE(), &pfd, flags, ResolveFlag::empty())?;
 
     // Record the file information,
     // so we can answer reliably on fstat.
@@ -1075,7 +1150,7 @@ fn handle_crypt(
     // SAFETY:
     // No need to wait on this lock, as we've just opened the file.
     // Close the send fd to get rid off our copy of the lock.
-    lock_fd(&aes_fd, false, false)?;
+    retry_on_eintr(|| lock_fd(&aes_fd, false, false))?;
     request.send_fd(aes_fd, o_cloexec, o_rand_fd)?;
 
     // Record encryption process information.
@@ -1119,12 +1194,19 @@ fn handle_creat(
     let req = request.scmpreq;
     let mask = proc_umask(req.pid())?;
 
+    // `path.base` may be empty for O_TMPFILE.
+    let base = if !path.base.is_empty() {
+        path.base
+    } else {
+        XPath::from_bytes(b".")
+    };
+
     // Set umask which is per-thread here.
     umask(mask);
 
-    // All set, make the open call.
+    // All set, make the openat2(2) call.
     let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
-    safe_openat2(fd, path.base, how)
+    safe_openat2(fd, base, how)
 }
 
 fn to_oflag(arg: u64, truncate: bool) -> Result<OFlag, Errno> {
@@ -1158,10 +1240,13 @@ fn to_oflag(arg: u64, truncate: bool) -> Result<OFlag, Errno> {
 }
 
 #[inline]
-fn to_mode(arg: u64, flags: OFlag) -> Result<Mode, Errno> {
+fn to_mode(arg: u64, flags: OFlag, ignore: bool) -> Result<Mode, Errno> {
     if flags.contains(OFlag::O_CREAT) || flags.contains(OFlag::O_TMPFILE) {
         // SAFETY: Reject invalid mode with creation flags.
         crate::kernel::to_mode(arg)
+    } else if !ignore && arg != 0 {
+        // SAFETY: mode argument must be zero without creation flags for openat2.
+        Err(Errno::EINVAL)
     } else {
         // SAFETY: mode argument is ignored without creation flags.
         // glibc sets it to 0, musl does not. Both are valid behaviour.
