@@ -8,37 +8,17 @@
 
 use std::os::fd::AsFd;
 
-use bitflags::bitflags;
 use libseccomp::ScmpNotifResp;
 use nix::errno::Errno;
 
 use crate::{
+    compat::RenameFlags,
     cookie::safe_renameat2,
     fs::FsFlags,
     hook::{PathArgs, SysArg, UNotifyEventRequest},
     kernel::syscall_path_handler,
 };
 
-// nix does not define RenameFlags for musl.
-bitflags! {
-    /// Flags for use with `renameat2`.
-    #[derive(Copy, Clone, Debug, Eq, Hash, Ord, PartialEq, PartialOrd)]
-    #[repr(transparent)]
-    pub(crate) struct RenameFlags: u32 {
-        /// Don’t overwrite `new_path` of the rename;
-        /// return an error if `new_path` already exists.
-        const RENAME_NOREPLACE = 1;
-
-        /// Atomically exchange `old_path` and `new_path`.
-        /// Both paths must exist.
-        const RENAME_EXCHANGE  = 2;
-
-        /// Create a “whiteout” at the source of the rename
-        /// (for overlay/union filesystems).
-        const RENAME_WHITEOUT  = 4;
-    }
-}
-
 pub(crate) fn sys_rename(request: UNotifyEventRequest) -> ScmpNotifResp {
     let argv = &[
         SysArg {
@@ -55,16 +35,11 @@ pub(crate) fn sys_rename(request: UNotifyEventRequest) -> ScmpNotifResp {
         },
     ];
 
-    syscall_path_handler(
-        request,
-        "rename",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "rename", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            syscall_rename_handler(request, path_args, RenameFlags::empty())
-        },
-    )
+        syscall_rename_handler(request, path_args, RenameFlags::empty())
+    })
 }
 
 pub(crate) fn sys_renameat(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -85,16 +60,11 @@ pub(crate) fn sys_renameat(request: UNotifyEventRequest) -> ScmpNotifResp {
         },
     ];
 
-    syscall_path_handler(
-        request,
-        "renameat",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "renameat", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            syscall_rename_handler(request, path_args, RenameFlags::empty())
-        },
-    )
+        syscall_rename_handler(request, path_args, RenameFlags::empty())
+    })
 }
 
 pub(crate) fn sys_renameat2(request: UNotifyEventRequest) -> ScmpNotifResp {
@@ -105,7 +75,14 @@ pub(crate) fn sys_renameat2(request: UNotifyEventRequest) -> ScmpNotifResp {
         Ok(flags) => flags,
         Err(errno) => return request.fail_syscall(errno),
     };
-    let noreplace = flags.contains(RenameFlags::RENAME_NOREPLACE);
+
+    // Note, (EXCHANGE | NOREPLACE) flags are mutually exclusive.
+    let mut fsflags_new = FsFlags::NO_FOLLOW_LAST | FsFlags::WANT_BASE;
+    if flags.contains(RenameFlags::RENAME_EXCHANGE) {
+        fsflags_new.insert(FsFlags::MUST_PATH);
+    } else if flags.contains(RenameFlags::RENAME_NOREPLACE) {
+        fsflags_new.insert(FsFlags::MISS_LAST);
+    }
 
     let argv = &[
         SysArg {
@@ -119,25 +96,16 @@ pub(crate) fn sys_renameat2(request: UNotifyEventRequest) -> ScmpNotifResp {
             dirfd: Some(2),
             path: Some(3),
             dotlast: Some(Errno::EINVAL),
-            fsflags: if noreplace {
-                FsFlags::MISS_LAST | FsFlags::NO_FOLLOW_LAST | FsFlags::WANT_BASE
-            } else {
-                FsFlags::NO_FOLLOW_LAST | FsFlags::WANT_BASE
-            },
+            fsflags: fsflags_new,
             ..Default::default()
         },
     ];
 
-    syscall_path_handler(
-        request,
-        "renameat2",
-        argv,
-        |path_args: PathArgs, request, sandbox| {
-            drop(sandbox); // release the read-lock.
+    syscall_path_handler(request, "renameat2", argv, |path_args, request, sandbox| {
+        drop(sandbox); // release the read-lock.
 
-            syscall_rename_handler(request, path_args, flags)
-        },
-    )
+        syscall_rename_handler(request, path_args, flags)
+    })
 }
 
 /// A helper function to handle rename and renameat syscalls.
@@ -173,5 +141,14 @@ fn syscall_rename_handler(
 #[inline]
 fn to_renameflags(arg: u64) -> Result<RenameFlags, Errno> {
     let flags = arg.try_into().or(Err(Errno::EINVAL))?;
-    RenameFlags::from_bits(flags).ok_or(Errno::EINVAL)
+    let flags = RenameFlags::from_bits(flags).ok_or(Errno::EINVAL)?;
+
+    // Fail if (NOREPLACE | WHITEOUT) is combined with EXCHANGE.
+    if flags.contains(RenameFlags::RENAME_EXCHANGE)
+        && flags.intersects(RenameFlags::RENAME_NOREPLACE | RenameFlags::RENAME_WHITEOUT)
+    {
+        return Err(Errno::EINVAL);
+    }
+
+    Ok(flags)
 }
