@@ -6,6 +6,7 @@
 //
 // SPDX-License-Identifier: GPL-3.0
 
+use bitflags::bitflags;
 use libseccomp::ScmpNotifResp;
 use nix::{
     errno::Errno,
@@ -14,30 +15,75 @@ use nix::{
 
 use crate::hook::UNotifyEventRequest;
 
+bitflags! {
+    // Bitflags describing signal system calls
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Hash)]
+    struct SignalSyscall: u8 {
+        // True if signal is group directed.
+        const SIGSYS_GROUP = 1 << 0;
+        // True if signal is thread directed.
+        const SIGSYS_THREAD = 1 << 1;
+        // True if signal is queued.
+        const SIGSYS_QUEUE = 1 << 2;
+    }
+}
+
+impl SignalSyscall {
+    fn is_group(self) -> bool {
+        self.contains(Self::SIGSYS_GROUP)
+    }
+
+    fn is_thread(self) -> bool {
+        self.contains(Self::SIGSYS_THREAD)
+    }
+
+    fn is_thread_group(self) -> bool {
+        self.contains(Self::SIGSYS_GROUP | Self::SIGSYS_THREAD)
+    }
+
+    fn is_queue(self) -> bool {
+        self.contains(Self::SIGSYS_QUEUE)
+    }
+}
+
 pub(crate) fn sys_kill(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_signal_handler(request, false, false)
+    syscall_signal_handler(request, SignalSyscall::empty())
 }
 
 pub(crate) fn sys_tgkill(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_signal_handler(request, true, true)
+    syscall_signal_handler(
+        request,
+        SignalSyscall::SIGSYS_GROUP | SignalSyscall::SIGSYS_THREAD,
+    )
 }
 
 pub(crate) fn sys_tkill(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_signal_handler(request, true, false)
+    syscall_signal_handler(request, SignalSyscall::SIGSYS_THREAD)
+}
+
+pub(crate) fn sys_rt_sigqueueinfo(request: UNotifyEventRequest) -> ScmpNotifResp {
+    syscall_signal_handler(request, SignalSyscall::SIGSYS_QUEUE)
+}
+
+pub(crate) fn sys_rt_tgsigqueueinfo(request: UNotifyEventRequest) -> ScmpNotifResp {
+    syscall_signal_handler(
+        request,
+        SignalSyscall::SIGSYS_QUEUE | SignalSyscall::SIGSYS_GROUP | SignalSyscall::SIGSYS_THREAD,
+    )
 }
 
 pub(crate) fn sys_pidfd_open(request: UNotifyEventRequest) -> ScmpNotifResp {
-    syscall_signal_handler(request, true, false)
+    // TODO: Recognize PIDFD_THREAD flag.
+    syscall_signal_handler(request, SignalSyscall::SIGSYS_THREAD)
 }
 
-// Maximum PID (wrap‐around limit) supported by the kernel.
+// Maximum PID (wrap-around limit) supported by the kernel.
 //
 // On 32-bit platforms this is fixed at 32768.
-// On 64-bit platforms it can go up to 2²² (approximately 4 million).
-#[cfg(target_pointer_width = "32")]
-const PID_MAX_LIMIT: libc::pid_t = 32_768;
-
-#[cfg(target_pointer_width = "64")]
+// On 64-bit platforms it can go up to 2^22 (approximately 4 million).
+//
+// Note, Syd may be built as 32-bit on a 64-bit platform,
+// therefore we always use the 64-bit limit.
 const PID_MAX_LIMIT: libc::pid_t = 1 << 22;
 
 /// Handles syscalls related to signal handling, protecting the syd
@@ -50,14 +96,11 @@ const PID_MAX_LIMIT: libc::pid_t = 1 << 22;
 ///   than a process.
 /// - `group`: true if the system call has both progress group id and
 ///   process id (tgkill), false otherwise.
+/// - `queue`: true if the system call is sigqueueinfo(2) or rt_sigqueueinfo.
 ///
 /// - `ScmpNotifResp`: Response indicating the result of the syscall handling.
 #[allow(clippy::cognitive_complexity)]
-fn syscall_signal_handler(
-    request: UNotifyEventRequest,
-    thread: bool,
-    group: bool,
-) -> ScmpNotifResp {
+fn syscall_signal_handler(request: UNotifyEventRequest, syscall: SignalSyscall) -> ScmpNotifResp {
     let req = request.scmpreq;
 
     // Validate pid/tid.
@@ -75,29 +118,37 @@ fn syscall_signal_handler(
         return request.fail_syscall(Errno::ESRCH);
     }
 
-    if group && tid == i32::MIN {
+    if syscall.is_group() && tid == i32::MIN {
         return request.fail_syscall(Errno::ESRCH);
     }
 
-    if !(-PID_MAX_LIMIT..=PID_MAX_LIMIT).contains(&pid) {
+    // kill(2) allows negative PIDs.
+    if syscall.is_empty() && !(-PID_MAX_LIMIT..=PID_MAX_LIMIT).contains(&pid) {
         return request.fail_syscall(Errno::ESRCH);
     }
 
-    if group && !(-PID_MAX_LIMIT..=PID_MAX_LIMIT).contains(&tid) {
-        return request.fail_syscall(Errno::ESRCH);
+    // rt_{tg,}sigqueueinfo return ESRCH/EINVAL on negative TGID.
+    if syscall.is_queue() && !(1..=PID_MAX_LIMIT).contains(&pid) {
+        return request.fail_syscall(if syscall.is_group() {
+            Errno::EINVAL
+        } else {
+            Errno::ESRCH
+        });
     }
 
-    if thread && (pid <= 0 || (group && tid <= 0)) {
+    // tkill(2), tgkill(2) and pidfd_open(2) return EINVAL on negative TGID.
+    if syscall.is_thread() && !(1..=PID_MAX_LIMIT).contains(&pid) {
+        return request.fail_syscall(Errno::EINVAL);
+    }
+
+    // tgkill(2) returns EINVAL on negative TID.
+    if syscall.is_group() && !(1..=PID_MAX_LIMIT).contains(&tid) {
         return request.fail_syscall(Errno::EINVAL);
     }
 
     // Guard syd tasks.
     //
-    // SAFETY: Return success when denying for stealth.
-    // Otherwise the allowed 0 signal can be misused
-    // to identify a Syd process.
-    //
-    // pid <=0 only for kill/sigqueue here.
+    // pid <=0 only for kill here.
     match pid {
         0 => {
             // SAFETY: Guard against group signals.
@@ -127,7 +178,7 @@ fn syscall_signal_handler(
     }
 
     // kill and sigqueue support negative PIDs.
-    let pid_abs = if thread { pid } else { pid.abs() };
+    let pid_abs = if syscall.is_thread() { pid } else { pid.abs() };
 
     // Check for Syd tasks.
     let syd = Pid::this().as_raw();
@@ -137,31 +188,33 @@ fn syscall_signal_handler(
     // respective signaler system call _only when_ the signal is 0.
     // Therefore we're not leaking any further information by returning
     // EACCES here.
-    if !thread && syd == pid_abs {
+    if !syscall.is_thread() && syd == pid_abs {
         return request.fail_syscall(Errno::EACCES);
     }
 
-    if thread && syd == pid {
+    if syscall.is_thread() && syd == pid {
         return request.fail_syscall(Errno::EACCES);
     }
 
-    if thread && group && syd == tid {
+    if syscall.is_thread_group() && syd == tid {
         return request.fail_syscall(Errno::EACCES);
     }
 
     // SAFETY: Check for Syd threads with the abstract PID.
-    if !thread && Errno::result(unsafe { libc::syscall(libc::SYS_tgkill, syd, pid_abs, 0) }).is_ok()
+    if !syscall.is_thread()
+        && Errno::result(unsafe { libc::syscall(libc::SYS_tgkill, syd, pid_abs, 0) }).is_ok()
     {
         return request.fail_syscall(Errno::EACCES);
     }
 
     // SAFETY: Check for Syd threads with the PID.
-    if thread && Errno::result(unsafe { libc::syscall(libc::SYS_tgkill, syd, pid, 0) }).is_ok() {
+    if syscall.is_thread()
+        && Errno::result(unsafe { libc::syscall(libc::SYS_tgkill, syd, pid, 0) }).is_ok()
+    {
         return request.fail_syscall(Errno::EACCES);
     }
 
-    if thread
-        && group
+    if syscall.is_thread_group()
         && pid != tid
         && Errno::result(
             // SAFETY: Check for Syd threads with the TID.
@@ -174,13 +227,13 @@ fn syscall_signal_handler(
 
     // Check signals directed to Syd's process group.
     let syd_pgid = getpgrp().as_raw();
-    if !thread && syd_pgid == pid_abs {
+    if !syscall.is_thread() && syd_pgid == pid_abs {
         return request.fail_syscall(Errno::EACCES);
     }
-    if thread && syd_pgid == pid {
+    if syscall.is_thread() && syd_pgid == pid {
         return request.fail_syscall(Errno::EACCES);
     }
-    if thread && group && syd_pgid == tid {
+    if syscall.is_thread_group() && syd_pgid == tid {
         return request.fail_syscall(Errno::EACCES);
     }
 
