@@ -7,32 +7,34 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
-    collections::HashSet,
     fs::File,
     io::BufReader,
-    os::{
-        fd::{AsFd, AsRawFd},
-        unix::ffi::OsStrExt,
-    },
+    os::fd::{AsFd, AsRawFd},
 };
 
 use libseccomp::ScmpNotifResp;
-use nix::{errno::Errno, NixPath};
+use nix::{errno::Errno, fcntl::AtFlags, NixPath};
 
 use crate::{
     compat::{fstatat64, statx, STATX_BASIC_STATS, STATX_MODE, STATX_TYPE},
     config::{MAGIC_PREFIX, MMAP_MIN_ADDR},
+    confine::{scmp_arch_bits, EOWNERDEAD},
     fs::{is_sidechannel_device, parse_fd, CanonicalPath, FileInfo, FileType, FsFlags},
+    hash::SydHashSet,
     hook::{SysArg, SysFlags, UNotifyEventRequest},
-    kernel::sandbox_path,
+    kernel::{sandbox_path, to_atflags},
     path::XPath,
     sandbox::Capability,
-    scmp_arch_bits,
 };
 
 pub(crate) fn sys_stat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[1] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let is32 = scmp_arch_bits(req.data.arch) == 32;
 
     let arg = SysArg {
@@ -45,6 +47,13 @@ pub(crate) fn sys_stat(request: UNotifyEventRequest) -> ScmpNotifResp {
 }
 
 pub(crate) fn sys_stat64(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[1] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let arg = SysArg {
         path: Some(0),
         fsflags: FsFlags::MUST_PATH,
@@ -57,6 +66,11 @@ pub(crate) fn sys_stat64(request: UNotifyEventRequest) -> ScmpNotifResp {
 pub(crate) fn sys_fstat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[1] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let is32 = scmp_arch_bits(req.data.arch) == 32;
 
     let arg = SysArg {
@@ -68,6 +82,13 @@ pub(crate) fn sys_fstat(request: UNotifyEventRequest) -> ScmpNotifResp {
 }
 
 pub(crate) fn sys_fstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[1] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let arg = SysArg {
         dirfd: Some(0),
         ..Default::default()
@@ -79,6 +100,11 @@ pub(crate) fn sys_fstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
 pub(crate) fn sys_lstat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[1] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let is32 = scmp_arch_bits(req.data.arch) == 32;
 
     let arg = SysArg {
@@ -91,6 +117,13 @@ pub(crate) fn sys_lstat(request: UNotifyEventRequest) -> ScmpNotifResp {
 }
 
 pub(crate) fn sys_lstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
+    let req = request.scmpreq;
+
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[1] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
     let arg = SysArg {
         path: Some(0),
         fsflags: FsFlags::MUST_PATH | FsFlags::NO_FOLLOW_LAST,
@@ -103,17 +136,33 @@ pub(crate) fn sys_lstat64(request: UNotifyEventRequest) -> ScmpNotifResp {
 pub(crate) fn sys_statx(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
-    let empty = req.data.args[2] & libc::AT_EMPTY_PATH as u64 != 0;
-    let follow = req.data.args[2] & libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[4] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
+    }
+
+    // SAFETY: Reject undefined/invalid flags.
+    // const AT_STATX_SYNC_AS_STAT: AtFlags = AtFlags::empty();
+    const AT_STATX_FORCE_SYNC: AtFlags = AtFlags::from_bits_retain(0x2000);
+    const AT_STATX_DONT_SYNC: AtFlags = AtFlags::from_bits_retain(0x4000);
+    let atflags = match to_atflags(
+        req.data.args[2],
+        AtFlags::AT_EMPTY_PATH
+            | AtFlags::AT_SYMLINK_NOFOLLOW
+            | AtFlags::AT_NO_AUTOMOUNT
+            | AT_STATX_FORCE_SYNC
+            | AT_STATX_DONT_SYNC,
+    ) {
+        Ok(atflags) => atflags,
+        Err(errno) => return request.fail_syscall(errno),
+    };
 
     let mut flags = SysFlags::empty();
     let mut fsflags = FsFlags::MUST_PATH;
-
-    if empty {
+    if atflags.contains(AtFlags::AT_EMPTY_PATH) {
         flags |= SysFlags::EMPTY_PATH;
     }
-
-    if !follow {
+    if atflags.contains(AtFlags::AT_SYMLINK_NOFOLLOW) {
         fsflags |= FsFlags::NO_FOLLOW_LAST;
     }
 
@@ -131,16 +180,26 @@ pub(crate) fn sys_statx(request: UNotifyEventRequest) -> ScmpNotifResp {
 pub(crate) fn sys_newfstatat(request: UNotifyEventRequest) -> ScmpNotifResp {
     let req = request.scmpreq;
 
-    let empty = req.data.args[3] & libc::AT_EMPTY_PATH as u64 != 0;
-    let follow = req.data.args[3] & libc::AT_SYMLINK_NOFOLLOW as u64 == 0;
-    let mut flags = SysFlags::empty();
-    let mut fsflags = FsFlags::MUST_PATH;
-
-    if empty {
-        flags |= SysFlags::EMPTY_PATH;
+    // Return EFAULT here for invalid pointers.
+    if req.data.args[2] < *MMAP_MIN_ADDR {
+        return request.fail_syscall(Errno::EFAULT);
     }
 
-    if !follow {
+    // SAFETY: Reject undefined/invalid flags.
+    let atflags = match to_atflags(
+        req.data.args[3],
+        AtFlags::AT_EMPTY_PATH | AtFlags::AT_SYMLINK_NOFOLLOW,
+    ) {
+        Ok(atflags) => atflags,
+        Err(errno) => return request.fail_syscall(errno),
+    };
+
+    let mut flags = SysFlags::empty();
+    let mut fsflags = FsFlags::MUST_PATH;
+    if atflags.contains(AtFlags::AT_EMPTY_PATH) {
+        flags |= SysFlags::EMPTY_PATH;
+    }
+    if atflags.contains(AtFlags::AT_SYMLINK_NOFOLLOW) {
         fsflags |= FsFlags::NO_FOLLOW_LAST;
     }
 
@@ -163,10 +222,6 @@ fn syscall_stat_handler(
     is32: bool,
 ) -> ScmpNotifResp {
     syscall_handler!(request, |request: UNotifyEventRequest| {
-        // Note: This is a virtual call handler,
-        // `sandbox` is an upgradable read lock with exclusive access.
-        // We'll either upgrade it or downgrade it based on magic lock.
-        // Exception: Sandbox lock had been set and there's no turning back.
         let req = request.scmpreq;
         let is_fd = arg.path.is_none();
         let sandbox = request.get_sandbox();
@@ -184,14 +239,8 @@ fn syscall_stat_handler(
         // If lock is on do not check for magic path.
         let (mut path, magic) = request.read_path(&sandbox, arg, !is_lock)?;
 
-        // SAFETY: For magic calls we allow NULL as stat argument, see syd(2).
-        // For other calls, return EFAULT here for invalid pointers.
-        let is_magic = !is_lock && magic;
-        if !is_magic && req.data.args[arg_stat] < *MMAP_MIN_ADDR {
-            return Err(Errno::EFAULT);
-        }
-
-        if is_magic {
+        let mut ghost = false;
+        if !is_lock && magic {
             drop(sandbox); // release the read-lock.
 
             // Handle magic prefix (ie /dev/syd)
@@ -209,20 +258,14 @@ fn syscall_stat_handler(
             let mut sandbox = request.get_mut_sandbox();
 
             // Execute magic command.
-            match cmd.as_os_str().as_bytes() {
-                b"ghost" => {
-                    // SAFETY: Reset sandbox to ensure no run-away execs.
-                    sandbox.reset()?;
-
-                    // Signal the poll process to exit.
-                    return Err(Errno::EOWNERDEAD);
-                }
-                b"panic" => sandbox.panic()?,
-                _ => {}
-            }
-
             if cmd.is_empty() || cmd.is_equal(b".el") || cmd.is_equal(b".sh") {
                 sandbox.config("")?;
+            } else if cmd.is_equal(b"panic") {
+                sandbox.panic()?;
+            } else if cmd.is_equal(b"ghost") {
+                // SAFETY: Reset sandbox to ensure no run-away execs.
+                sandbox.reset()?;
+                ghost = true;
             } else if let Some(cmd) = cmd.strip_prefix(b"load") {
                 // We handle load specially here as it involves process access.
                 // 1. Attempt to parse as FD, pidfd_getfd and load it.
@@ -231,7 +274,7 @@ fn syscall_stat_handler(
                     Ok(remote_fd) => {
                         let fd = request.get_fd(remote_fd)?;
                         let file = BufReader::new(File::from(fd));
-                        let mut imap = HashSet::default();
+                        let mut imap = SydHashSet::default();
                         // SAFETY: parse_config() checks for the file name
                         // /dev/syd/load and disables config file include
                         // feature depending on this check.
@@ -261,11 +304,6 @@ fn syscall_stat_handler(
                 return Err(Errno::EINVAL);
             }
             drop(sandbox); // release the write-lock.
-
-            // If the stat buffer is NULL, return immediately.
-            if req.data.args[arg_stat] == 0 {
-                return Ok(request.return_syscall(0));
-            }
         } else {
             // Handle fstat for files with encryption in progress.
             let mut crypt_stat = false;
@@ -320,20 +358,20 @@ fn syscall_stat_handler(
         //
         // Return ENOTDIR for non-directories with trailing slash.
         if let Some(file_type) = &path.typ {
-            if !matches!(file_type, FileType::Dir | FileType::MagicLnk(_, _))
+            if !matches!(file_type, FileType::Dir | FileType::MagicLnk(_))
                 && path.abs().last() == Some(b'/')
             {
                 return Err(Errno::ENOTDIR);
             }
         }
 
-        let mut flags = if path.base.is_empty() {
-            libc::AT_EMPTY_PATH
-        } else {
-            // SAFETY: After this point we are not permitted to resolve
-            // symbolic links any longer or else we risk TOCTOU.
-            libc::AT_SYMLINK_NOFOLLOW
-        };
+        // We use MUST_PATH, dir refers to the file.
+        assert!(
+            path.base.is_empty(),
+            "BUG: MUST_PATH returned a directory for stat, report a bug!"
+        );
+        let fd = path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?;
+        let mut flags = libc::AT_EMPTY_PATH;
 
         #[allow(clippy::cast_possible_truncation)]
         if arg_stat == 4 {
@@ -365,12 +403,7 @@ fn syscall_stat_handler(
             }
 
             // Note, unlike statfs, stat does not EINTR.
-            let mut statx = statx(
-                path.dir.as_ref().map(|fd| fd.as_fd()).ok_or(Errno::EBADF)?,
-                path.base,
-                flags,
-                mask,
-            )?;
+            let mut statx = statx(fd, "", flags, mask)?;
 
             // SAFETY: Check if the file is a sidechannel device and
             // update its access and modification times to match the
@@ -418,7 +451,7 @@ fn syscall_stat_handler(
 
             // SAFETY: In libc we trust.
             // Note, unlike statfs, stat does not EINTR.
-            let mut stat = fstatat64(path.dir.as_ref().map(|fd| fd.as_raw_fd()), path.base, flags)?;
+            let mut stat = fstatat64(Some(fd.as_raw_fd()), "", flags)?;
 
             // SAFETY: Check if the file is a sidechannel device and
             // update its access and modification times to match the
@@ -475,7 +508,14 @@ fn syscall_stat_handler(
             }
         }
 
-        // stat system call successfully emulated.
+        // Use the pseudo errno(3) EOWNERDEAD to initiate ghost mode.
+        // We only do it here to ensure metadata of /dev/null was
+        // written to sandbox process memory.
+        if ghost {
+            return Ok(ScmpNotifResp::new(0, 0, EOWNERDEAD, 0));
+        }
+
+        // stat(2) system call has been successfully emulated.
         Ok(request.return_syscall(0))
     })
 }
