@@ -69,9 +69,8 @@
 //! from `/proc/$pid/mem`. Linux-5.11 or later is recommended.
 
 // We like clean and simple code with documentation.
-// Keep in sync with main.rs.
+// Keep in sync with syd.rs.
 #![deny(missing_docs)]
-#![deny(clippy::allow_attributes_without_reason)]
 #![deny(clippy::arithmetic_side_effects)]
 #![deny(clippy::as_ptr_cast_mut)]
 #![deny(clippy::as_underscore)]
@@ -113,7 +112,7 @@ pub mod api;
 #[cfg(feature = "asm")]
 pub mod asm;
 /// System call argument cookies
-pub(crate) mod cookie;
+pub mod cookie;
 /// Utilities to mask sensitive information in proc files
 pub(crate) mod mask;
 /// System call handlers
@@ -168,7 +167,7 @@ pub mod seal;
 pub mod sealbox;
 /// Portable sigset that can handle reserved signals
 pub mod sigset;
-/// Interface to Linux prctl(2) speculation misfeature interfac
+/// Interface to Linux prctl(2) speculation misfeature interface
 pub mod spec;
 /// sysinfo(2) interface
 pub mod sysinfo;
@@ -235,15 +234,16 @@ use nix::{
     },
     unistd::{Gid, Group, Uid, User},
 };
+use serde::{Serialize, Serializer};
 
 use crate::{
     compat::{fstatx, lsm_list_modules},
     confine::{is_coredump, lock_enabled, ns_enabled, seccomp_arch_native_name, SydPersona},
     err::err2no,
-    hash::{aes_ctr_info, hmac_sha256_info},
+    hash::{aes_ctr_info, check_setsockopt_serial_support, hmac_sha256_info, key_ring_validate},
     landlock::ABI,
     path::{XPath, XPathBuf},
-    proc::{proc_fs_file_max, proc_fs_nr_open, proc_kernel_randomize_va_space},
+    proc::{proc_fs_file_max, proc_fs_nr_open, proc_kernel_randomize_va_space, proc_kernel_taint},
     sealbox::check_mseal_support,
     spec::{speculation_get, SpeculationFeature},
 };
@@ -295,38 +295,13 @@ pub fn check_cross_memory_attach() -> bool {
 /// and information about the system to
 /// standard output.
 #[allow(clippy::cognitive_complexity)]
-pub fn syd_info() {
+pub fn syd_info(verbose: bool) -> Result<(), Errno> {
     use crate::config::*;
 
-    #[allow(clippy::disallowed_methods)]
-    let major = env!("CARGO_PKG_VERSION_MAJOR")
-        .parse::<u64>()
-        .expect("CARGO_PKG_VERSION_MAJOR");
-    #[allow(clippy::disallowed_methods)]
-    let minor = env!("CARGO_PKG_VERSION_MINOR")
-        .parse::<u64>()
-        .expect("CARGO_PKG_VERSION_MINOR");
-    #[allow(clippy::disallowed_methods)]
-    let patch = env!("CARGO_PKG_VERSION_PATCH")
-        .parse::<u64>()
-        .expect("CARGO_PKG_VERSION_PATCH");
-    let hex_version = (major << 16) | (minor << 8) | patch;
-    let code_name = hex_version
-        .to_name()
-        .split('_')
-        .map(|word| {
-            let mut c = word.chars();
-            match c.next() {
-                None => String::new(),
-                Some(f) => f.to_uppercase().collect::<String>() + c.as_str(),
-            }
-        })
-        .collect::<Vec<String>>()
-        .join(" ");
-
-    println!("syd {} ({})", *crate::config::VERSION, code_name);
-    println!("Author: Ali Polatel <alip@chesswob.org>");
-    println!("License: GPL-3.0");
+    printfln!("syd {} ({})", *crate::config::VERSION, syd_code_name())?;
+    printfln!("Rock solid application kernel")?;
+    printfln!("Author: Ali Polatel <alip@chesswob.org>")?;
+    printfln!("License: GPL-3.0-only")?;
 
     let feat = [
         #[cfg(debug_assertions)]
@@ -346,21 +321,33 @@ pub fn syd_info() {
         #[cfg(not(feature = "prof"))]
         "-prof",
     ];
-    println!("Features: {}", feat.join(", "));
+    printfln!("Features: {}", feat.join(", "))?;
+
+    if !verbose {
+        return Ok(());
+    }
 
     let libapi = libseccomp::get_api();
     match ScmpVersion::current() {
         Ok(libver) => {
-            println!(
+            printfln!(
                 "LibSeccomp: v{}.{}.{} api:{}",
-                libver.major, libver.minor, libver.micro, libapi
-            );
+                libver.major,
+                libver.minor,
+                libver.micro,
+                libapi
+            )?;
         }
         Err(error) => {
-            println!("LibSeccomp: ? (error: {error})");
+            printfln!("LibSeccomp: ? (error: {error})")?;
         }
     }
 
+    match proc_kernel_taint() {
+        Ok(tflags) => printfln!("{tflags}"),
+        Err(errno) => printfln!("Kernel may be tainted (error: {errno})."),
+    }?;
+
     let aslr = match proc_kernel_randomize_va_space() {
         Ok(0) => "disabled".to_string(),
         Ok(1) => "enabled (stack, mmap, VDSO; PIE text randomized)".to_string(),
@@ -368,7 +355,7 @@ pub fn syd_info() {
         Ok(n) => format!("{n} (error: {})", Errno::EINVAL),
         Err(errno) => format!("? (error: {errno})"),
     };
-    println!("ASLR is {aslr}.");
+    printfln!("ASLR is {aslr}.")?;
 
     #[allow(clippy::disallowed_methods)]
     let bpf_jit = match std::fs::read_to_string("/proc/sys/net/core/bpf_jit_enable") {
@@ -380,11 +367,11 @@ pub fn syd_info() {
         },
         Err(err) => format!("? (error: {})", err2no(&err)),
     };
-    println!("BPF JIT compiler is {bpf_jit}.");
+    printfln!("BPF JIT compiler is {bpf_jit}.")?;
 
     let abi = ABI::new_current();
     if abi == ABI::Unsupported {
-        println!("Landlock is not supported.");
+        printfln!("Landlock is not supported.")?;
     } else {
         let state = lock_enabled(abi);
         let state_verb = match state {
@@ -393,21 +380,21 @@ pub fn syd_info() {
             2 => "not enforced",
             _ => "unsupported",
         };
-        println!("Landlock ABI {} is {state_verb}.", abi as i32);
+        printfln!("Landlock ABI {} is {state_verb}.", abi as i32)?;
     }
 
-    println!(
+    printfln!(
         "User namespaces are {}supported.",
         if ns_enabled(CloneFlags::CLONE_NEWUSER).unwrap_or(false) {
             ""
         } else {
             "not "
         }
-    );
+    )?;
 
     // Check CONFIG_CROSS_MEMORY_ATTACH.
     let cfg_cma = check_cross_memory_attach();
-    println!(
+    printfln!(
         "Cross memory attach is {}supported{}",
         if cfg_cma { "" } else { "not " },
         if cfg_cma {
@@ -415,17 +402,37 @@ pub fn syd_info() {
         } else {
             " (\x1b[91minsecure\x1b[0m)."
         },
-    );
+    )?;
 
-    println!(
+    printfln!(
         "Memory sealing is {}supported.",
         if check_mseal_support() { "" } else { "not " }
-    );
+    )?;
+
+    // Print ALG_SET_KEY_BY_KEY_SERIAL support.
+    printfln!(
+        "Algorithm sockets {} keyrings(7) support.",
+        if check_setsockopt_serial_support() {
+            "have"
+        } else {
+            "doesn't have"
+        }
+    )?;
+
+    // Print whether session keyring is attached to the user keyring.
+    match key_ring_validate() {
+        Ok(()) => {
+            printfln!("Session keyring is attached to the user keyring.")?;
+        }
+        Err(errno) => {
+            printfln!("Session keyring isn't attached to the user keyring: {errno}!")?;
+        }
+    }
 
     // Print ctr(aes) kernel support.
-    println!("{}", aes_ctr_info());
+    printfln!("{}", aes_ctr_info())?;
     // Print hmac(sha256) kernel support.
-    println!("{}", hmac_sha256_info());
+    printfln!("{}", hmac_sha256_info())?;
 
     let lsms = match lsm_list_modules() {
         Ok(lsms) => lsms
@@ -436,54 +443,54 @@ pub fn syd_info() {
         Err(Errno::ENOENT) => "none loaded".to_string(),
         Err(errno) => format!("? (error: {errno})"),
     };
-    println!("LSMs: {lsms}.");
+    printfln!("LSMs: {lsms}.")?;
 
     // Print information on open file limits.
     let (nofile_soft, nofile_hard) = getrlimit(Resource::RLIMIT_NOFILE).unwrap_or((0, 0));
-    print!("Open file limits: {nofile_soft} soft, {nofile_hard} hard, ");
+    printf!("Open file limits: {nofile_soft} soft, {nofile_hard} hard, ")?;
 
     // Lookup system-wide open file limits.
     let file_max = proc_fs_file_max().unwrap_or(0);
     let nr_open = proc_fs_nr_open().unwrap_or(0);
-    println!("{nr_open} nr_open, {file_max} file-max");
+    printfln!("{nr_open} nr_open, {file_max} file-max")?;
 
     let uname = match uname() {
         Ok(info) => OsStr::to_str(info.release()).unwrap_or("?").to_string(),
         Err(_) => "?".to_string(),
     };
-    println!("Host (build): {}", env!("SYD_BUILDHOST"));
-    println!(
+    printfln!("Host (build): {}", env!("SYD_BUILDHOST"))?;
+    printfln!(
         "Host (target): {uname} {}",
         seccomp_arch_native_name().unwrap_or("?")
-    );
+    )?;
 
     // Print detected host Linux kernel version and related features.
-    print!("Host Linux: {}.{} with", KERNEL_VERSION.0, KERNEL_VERSION.1);
-    print!(
+    printf!("Host Linux: {}.{} with", KERNEL_VERSION.0, KERNEL_VERSION.1)?;
+    printf!(
         " {}pidfd_thread",
         if *HAVE_PIDFD_THREAD { "+" } else { "-" }
-    );
-    print!(
+    )?;
+    printf!(
         ", {}procmap_query",
         if *HAVE_PROCMAP_QUERY { "+" } else { "-" }
-    );
-    print!(
+    )?;
+    printf!(
         ", {}seccomp_user_notif_fd_sync_wake_up",
         if *HAVE_SECCOMP_USER_NOTIF_FD_SYNC_WAKE_UP {
             "+"
         } else {
             "-"
         }
-    );
-    print!(
+    )?;
+    printf!(
         ", {}statx_mnt_id_unique",
         if *HAVE_STATX_MNT_ID_UNIQUE { "+" } else { "-" }
-    );
-    print!(
+    )?;
+    printf!(
         ", {}at_execve_check",
         if *HAVE_AT_EXECVE_CHECK { "+" } else { "-" }
-    );
-    println!(", mmap_min_addr={}", *MMAP_MIN_ADDR);
+    )?;
+    printfln!(", mmap_min_addr={}", *MMAP_MIN_ADDR)?;
 
     // SAFETY: In libc we trust.
     // Note: nix version truncates unknown bits which we don't want.
@@ -492,33 +499,64 @@ pub fn syd_info() {
         Err(errno) => format!("? (error: {errno})"),
     };
 
-    println!(
+    printfln!(
         "Environment: {}-{pers}-{}",
         env!("SYD_TARGET_ENV"),
         env!("SYD_TARGET_POINTER_WIDTH")
-    );
+    )?;
 
-    println!(
+    printfln!(
         "CPU: {} ({} cores), {}-endian",
         num_cpus::get(),
         num_cpus::get_physical(),
         env!("SYD_TARGET_ENDIAN")
-    );
-    println!("CPUFLAGS: {}", env!("SYD_TARGET_FEATURE"));
+    )?;
+    printfln!("CPUFLAGS: {}", env!("SYD_TARGET_FEATURE"))?;
 
     for spec_feat in [
         SpeculationFeature::StoreBypass,
         SpeculationFeature::IndirectBranch,
         SpeculationFeature::L1DFlush,
     ] {
-        println!(
+        printfln!(
             "{}",
             match speculation_get(spec_feat) {
                 Ok(status) => status.to_string(),
                 Err(errno) => format!("{spec_feat} status: ? (error: {errno})"),
             }
-        );
+        )?;
     }
+
+    Ok(())
+}
+
+/// Print Syd code name.
+pub fn syd_code_name() -> String {
+    #[allow(clippy::disallowed_methods)]
+    let major = env!("CARGO_PKG_VERSION_MAJOR")
+        .parse::<u64>()
+        .expect("CARGO_PKG_VERSION_MAJOR");
+    #[allow(clippy::disallowed_methods)]
+    let minor = env!("CARGO_PKG_VERSION_MINOR")
+        .parse::<u64>()
+        .expect("CARGO_PKG_VERSION_MINOR");
+    #[allow(clippy::disallowed_methods)]
+    let patch = env!("CARGO_PKG_VERSION_PATCH")
+        .parse::<u64>()
+        .expect("CARGO_PKG_VERSION_PATCH");
+    let hex_version = (major << 16) | (minor << 8) | patch;
+    hex_version
+        .to_name()
+        .split('_')
+        .map(|word| {
+            let mut c = word.chars();
+            match c.next() {
+                None => String::new(),
+                Some(f) => f.to_uppercase().collect::<String>() + c.as_str(),
+            }
+        })
+        .collect::<Vec<String>>()
+        .join(" ")
 }
 
 /// Given a `Uid`, return the user name of the user.
@@ -572,26 +610,60 @@ pub fn reset_signal(signal: Signal) -> Result<(), Errno> {
     unsafe { sigaction(signal, &sig_action) }.map(drop)
 }
 
-/// Ignores all signals except SIG{KILL,STOP,PIPE,CHLD},
-/// and all signals with default action Core.
-pub fn ignore_signals() -> Result<(), Errno> {
+bitflags::bitflags! {
+    /// Represents valid `ignore_signals` options.
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
+    pub struct IgnoreSignalOpts: u8 {
+        /// Skip ignoring the `SIGALRM` signal.
+        const SkipIgnoreAlarm = 1 << 0;
+        /// Skip ignoring signals with default action Core.
+        const SkipIgnoreCoreDump = 1 << 1;
+    }
+}
+
+impl Serialize for IgnoreSignalOpts {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let mut opts: Vec<&str> = vec![];
+
+        if self.is_empty() {
+            return serializer.collect_seq(opts);
+        }
+        if self.contains(Self::SkipIgnoreAlarm) {
+            opts.push("skip_ignore_alarm");
+        }
+        if self.contains(Self::SkipIgnoreCoreDump) {
+            opts.push("skip_ignore_core_dump");
+        }
+
+        opts.sort();
+        serializer.collect_seq(opts)
+    }
+}
+
+/// Ignores all signals except SIG{ALRM,CHLD,KILL,STOP}.
+///
+/// Skips ignoring SIGPROF signal when profiling is enabled with the `prof` feature.
+/// Skips ignoring SIGALRM signal if `IgnoreSignalOpts::SkipIgnoreAlarm` is set.
+/// Skips ignoring signals with default action Core if `IgnoreSignalOpts::SkipIgnoreCoreDump` is set.
+pub fn ignore_signals(opts: IgnoreSignalOpts) -> Result<(), Errno> {
     // Iterate through all possible signals and set them to be ignored.
     // Step 1: Normal signals.
     for signal in Signal::iterator() {
         // 1. Can not ignore SIGKILL and SIGSTOP.
         // 2. Do not need to ignore Signals with default action Core.
         // 3. Ignoring SIGCHLD changes wait semantics which we cannot do.
-        // 4. SIGPIPE must be handled outside this function.
-        // 5. SIGPROF is used for profiling.
+        // 4. SIGPROF is used for profiling.
         match signal {
-            Signal::SIGALRM
-            | Signal::SIGKILL
-            | Signal::SIGSTOP
-            | Signal::SIGPIPE
-            | Signal::SIGCHLD => {}
+            Signal::SIGCHLD | Signal::SIGKILL | Signal::SIGSTOP => {}
             #[cfg(feature = "prof")]
             Signal::SIGPROF => {}
-            signal if is_coredump(signal as i32) => {}
+            Signal::SIGALRM if opts.contains(IgnoreSignalOpts::SkipIgnoreAlarm) => {}
+            signal
+                if opts.contains(IgnoreSignalOpts::SkipIgnoreCoreDump)
+                    && is_coredump(signal as i32) => {}
             signal => ignore_signal(signal)?,
         }
     }
