@@ -20,11 +20,12 @@ use std::{
         unix::ffi::{OsStrExt, OsStringExt},
     },
     path::{Path, PathBuf},
+    sync::LazyLock,
 };
 
 use btoi::btoi;
 use memchr::{
-    arch::all::{is_equal, is_prefix, is_suffix},
+    arch::all::{is_equal, is_prefix, is_suffix, memchr::One},
     memchr, memmem, memrchr,
 };
 use nix::{
@@ -34,11 +35,10 @@ use nix::{
     unistd::Pid,
     NixPath,
 };
-use once_cell::sync::Lazy;
 
 use crate::{
     config::MAGIC_PREFIX,
-    fs::{retry_on_eintr, FileType},
+    fs::{retry_on_eintr, tgkill, FileType},
     log::log_untrusted_buf,
 };
 
@@ -57,7 +57,7 @@ pub const PATH_MAX: usize = 4096;
 pub const PATH_MIN: usize = 64;
 
 // This pointer is confined by seccomp for use with openat(2) for getdir_long().
-static DOTDOT: Lazy<u64> = Lazy::new(|| c"..".as_ptr() as *const libc::c_char as u64);
+static DOTDOT: LazyLock<u64> = LazyLock::new(|| c"..".as_ptr() as *const libc::c_char as u64);
 
 #[inline(always)]
 pub(crate) fn dotdot_with_nul() -> u64 {
@@ -65,7 +65,7 @@ pub(crate) fn dotdot_with_nul() -> u64 {
 }
 
 /// `PathBuf` for UNIX.
-// SAFETY: k1 == k2 ⇒ hash(k1) == hash(k2) always holds for our PartialEq impl.
+// SAFETY: k1 == k2 -> hash(k1) == hash(k2) always holds for our PartialEq impl.
 #[allow(clippy::derived_hash_with_manual_eq)]
 #[derive(Clone, Hash, Ord, PartialOrd)]
 pub struct XPathBuf(Vec<u8>);
@@ -347,12 +347,16 @@ impl XPath {
         file_type: Option<&FileType>,
         dir_entry: Option<&XPath>,
         safe_name: bool,
+        restrict_mkbdev: bool,
     ) -> Result<(), Errno> {
         //
         // RESTRICTION 1: Prevent listing block devices and files of unknown type.
         //
         // SAFETY: Prevent listing block devices and files of unknown type.
-        if matches!(file_type, Some(FileType::Blk | FileType::Unk)) {
+        // Block device restriction may be relaxed with trace/allow_unsafe_mkbdev:1.
+        if file_type == Some(&FileType::Unk)
+            || (restrict_mkbdev && file_type == Some(&FileType::Blk))
+        {
             return Err(Errno::ENOENT);
         }
         // END OF RESTRICTION 1
@@ -421,15 +425,14 @@ impl XPath {
         // SAFETY: Protect Syd /proc directory!
         //
         // Step 1: Protect Syd thread group.
-        let syd_pid = Pid::this().as_raw();
+        let syd_pid = Pid::this();
+        let proc_pid = Pid::from_raw(proc_pid);
         if proc_pid == syd_pid {
             return Err(Errno::ENOENT);
         }
         //
         // Step 2: Protect all Syd threads.
-        //
-        // SAFETY: In libc we trust.
-        if unsafe { libc::syscall(nix::libc::SYS_tgkill, syd_pid, proc_pid, 0) } == 0 {
+        if tgkill(syd_pid, proc_pid, 0).is_ok() {
             return Err(Errno::ENOENT);
         }
         // END OF RESTRICTION 3
@@ -492,7 +495,24 @@ impl XPath {
         }
 
         // Check if the filename is valid UTF-8.
-        if std::str::from_utf8(name).is_err() {
+        let name_utf8 = std::str::from_utf8(name).or(Err(Errno::EINVAL))?;
+
+        // Check if first and last character is not whitespace.
+        // This includes UTF-8 whitespace.
+        if name_utf8
+            .chars()
+            .nth(0)
+            .map(|c| c.is_whitespace())
+            .unwrap_or(false)
+        {
+            return Err(Errno::EINVAL);
+        }
+        if name_utf8
+            .chars()
+            .last()
+            .map(|c| c.is_whitespace())
+            .unwrap_or(false)
+        {
             return Err(Errno::EINVAL);
         }
 
@@ -784,7 +804,14 @@ impl XPath {
         self.starts_with(MAGIC_PREFIX)
     }
 
+    /// Check if path is literally the root path, ie `/`.
+    pub fn is_root(&self) -> bool {
+        self.is_equal(b"/")
+    }
+
     /// Check if path is the root path, ie `/`.
+    ///
+    /// Path may consist of multiple slashes and single dots.
     pub fn is_rootfs(&self) -> bool {
         self.as_bytes().iter().all(|b| matches!(*b, b'/' | b'.'))
     }
@@ -831,10 +858,28 @@ impl XPath {
         if !self.is_proc() {
             return false;
         }
-        match self.get("/proc/".len()) {
-            Some(n) => n.is_ascii_digit(),
-            None => false,
+        self.get("/proc/".len())
+            .map(|b| b.is_ascii_digit())
+            .unwrap_or(false)
+    }
+
+    /// Check if path points to machine-id(5), /etc/hostid or /var/adm/hostid.
+    pub fn is_machine_id(&self) -> bool {
+        for f in [
+            &b"/etc/machine-id"[..],
+            &b"/etc/hostid"[..],
+            &b"/var/adm/hostid"[..],
+        ] {
+            if self.is_equal(f) {
+                return true;
+            }
         }
+        false
+    }
+
+    /// Check if path points to a proc_pid_status(5) file.
+    pub fn is_proc_pid_status(&self) -> bool {
+        self.is_proc_pid() && self.ends_with(b"/status")
     }
 
     /// Check if path points to the `/proc/self` link.
@@ -905,11 +950,28 @@ impl XPath {
         }
 
         let mut has_component = false;
+        let mut start = 0;
 
-        for component in bytes.split(|&b| b == b'/') {
-            if component.is_empty() {
-                continue;
+        // Iterate over '/' positions without allocating.
+        #[allow(clippy::arithmetic_side_effects)]
+        for sep in One::new(b'/').iter(bytes) {
+            if sep > start {
+                // Non-empty component [start, sep).
+                let component = &bytes[start..sep];
+                has_component = true;
+                if component != b"." {
+                    return false;
+                }
             }
+
+            // Move past '/'.
+            start = sep + 1;
+        }
+
+        // Handle the trailing component after the last '/',
+        // or the whole slice if none.
+        if start < bytes.len() {
+            let component = &bytes[start..];
             has_component = true;
             if component != b"." {
                 return false;
@@ -936,7 +998,7 @@ impl XPath {
 
     /// Determine whether path contains the given substring.
     pub fn contains(&self, sub: &[u8]) -> bool {
-        memmem::find_iter(self.as_bytes(), sub).next().is_some()
+        memmem::find(self.as_bytes(), sub).is_some()
     }
 
     /// Determine whether path contains the given character.
@@ -1065,6 +1127,70 @@ impl XPathBuf {
         self.0.truncate(write_pos);
     }
 
+    /// Replace the **leading** `old` prefix with `new` in-place.
+    ///
+    /// - No-op if the buffer does **not** start with `old`, or already starts with `new`,
+    ///   or if `old == new`.
+    ///
+    /// ### Errors
+    /// - `EINVAL` if `old` is empty.
+    /// - `EOVERFLOW` on arithmetic overflow when growing the buffer.
+    /// - `ENOMEM` if additional capacity is required and cannot be reserved.
+    ///
+    /// This performs at most one `copy_within` of the tail and one prefix overwrite; it
+    /// allocates only if `new.len() > old.len()`. Designed for cases like upgrading
+    /// `!memfd:` -> `!memfd-hugetlb:`.
+    pub fn replace_prefix(&mut self, old: &[u8], new: &[u8]) -> Result<(), Errno> {
+        if old.is_empty() {
+            return Err(Errno::EINVAL);
+        }
+
+        let s = self.as_bytes();
+
+        // If already has the desired prefix exactly, do nothing.
+        if is_prefix(s, new) || !is_prefix(s, old) || old == new {
+            return Ok(());
+        }
+
+        let old_len = old.len();
+        let new_len = new.len();
+        let len = self.0.len();
+
+        #[allow(clippy::arithmetic_side_effects)]
+        match new_len.cmp(&old_len) {
+            Ordering::Equal => {
+                // Same length: overwrite the prefix.
+                self.0[..new_len].copy_from_slice(new);
+            }
+            Ordering::Less => {
+                // Shrink: move tail left, then overwrite head.
+                let diff = old_len - new_len;
+                if old_len <= len {
+                    self.0.copy_within(old_len..len, new_len);
+                    let new_total = len.saturating_sub(diff);
+                    self.0.truncate(new_total);
+                    self.0[..new_len].copy_from_slice(new);
+                } else {
+                    // Defensive: should not happen if starts_with(old) was true.
+                    return Err(Errno::EINVAL);
+                }
+            }
+            Ordering::Greater => {
+                // Grow: reserve, extend, move tail right, then overwrite head.
+                let add = new_len - old_len;
+                let new_total = len.checked_add(add).ok_or(Errno::EOVERFLOW)?;
+                self.try_reserve(add)?;
+                self.0.resize(new_total, 0);
+                // Move previous tail (len - old_len bytes) right by `add`.
+                self.0
+                    .copy_within(old_len..(new_total - add), old_len + add);
+                self.0[..new_len].copy_from_slice(new);
+            }
+        }
+
+        Ok(())
+    }
+
     /// Create a path from the given PID.
     pub fn from_pid(pid: Pid) -> Self {
         let mut buf = itoa::Buffer::new();
@@ -1269,6 +1395,12 @@ impl XPathBuf {
         self.0.as_mut_ptr()
     }
 
+    /// Sets an index in path to the given value.
+    /// Panics if the index is invalid.
+    pub fn set(&mut self, idx: usize, val: u8) {
+        self.0[idx] = val;
+    }
+
     /// Forces the length of the internal `Vec`.
     ///
     /// # Safety
@@ -1306,9 +1438,12 @@ fn is_permitted_final(b: u8) -> bool {
 
 #[inline]
 fn is_permitted_byte(b: u8) -> bool {
+    // As of version 3.38.0 the following characters are removed:
+    // 1. `:` which is used commonly across devfs and procfs.
+    // 2. `{` and `}` which are used in Firefox profile directories.
     match b {
-        b'*' | b'?' | b':' | b'[' | b']' | b'"' | b'<' | b'>' | b'|' | b'(' | b')' | b'{'
-        | b'}' | b'&' | b'\'' | b'!' | b'\\' | b';' | b'$' | b'`' => false,
+        b'*' | b'?' | b'[' | b']' | b'"' | b'<' | b'>' | b'|' | b'(' | b')' | b'&' | b'\''
+        | b'!' | b'\\' | b';' | b'$' | b'`' => false,
         0x20..=0x7E => true,
         0x80..=0xFE => true,
         _ => false,
@@ -1500,7 +1635,7 @@ mod tests {
     #[test]
     fn test_path_check_file_type() {
         assert!(XPathBuf::from("/proc")
-            .check(Pid::from_raw(1), Some(&FileType::Dir), None, true)
+            .check(Pid::from_raw(1), Some(&FileType::Dir), None, true, true)
             .is_ok());
         assert!(XPathBuf::from("/proc")
             .check(
@@ -1508,6 +1643,7 @@ mod tests {
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"self")),
                 true,
+                true,
             )
             .is_ok());
         assert!(XPathBuf::from("/proc")
@@ -1516,22 +1652,23 @@ mod tests {
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"uptime")),
                 true,
+                true,
             )
             .is_ok());
         assert!(XPathBuf::from("/dev/null")
-            .check(Pid::from_raw(1), Some(&FileType::Chr), None, true)
+            .check(Pid::from_raw(1), Some(&FileType::Chr), None, true, true)
             .is_ok());
         assert!(XPathBuf::from("/dev/log")
-            .check(Pid::from_raw(1), Some(&FileType::Sock), None, true)
+            .check(Pid::from_raw(1), Some(&FileType::Sock), None, true, true)
             .is_ok());
         assert!(XPathBuf::from("/dev/fifo")
-            .check(Pid::from_raw(1), Some(&FileType::Fifo), None, true)
+            .check(Pid::from_raw(1), Some(&FileType::Fifo), None, true, true)
             .is_ok());
         assert!(XPathBuf::from("/dev/sda1")
-            .check(Pid::from_raw(1), Some(&FileType::Blk), None, true)
+            .check(Pid::from_raw(1), Some(&FileType::Blk), None, true, true)
             .is_err());
         assert!(XPathBuf::from("/dev/lmao")
-            .check(Pid::from_raw(1), Some(&FileType::Unk), None, true)
+            .check(Pid::from_raw(1), Some(&FileType::Unk), None, true, true)
             .is_err());
     }
 
@@ -1540,7 +1677,13 @@ mod tests {
         let this = Pid::from_raw(128);
         let that = Pid::from_raw(256);
         assert!(XPathBuf::from("/proc")
-            .check(this, Some(&FileType::Dir), Some(&xpath!("{this}")), true,)
+            .check(
+                this,
+                Some(&FileType::Dir),
+                Some(&xpath!("{this}")),
+                true,
+                true
+            )
             .is_ok());
         assert!(XPathBuf::from(format!("/proc/{this}"))
             .check(
@@ -1548,6 +1691,7 @@ mod tests {
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"mem")),
                 true,
+                true,
             )
             .is_ok());
         assert!(XPathBuf::from(format!("/proc/{this}"))
@@ -1556,13 +1700,26 @@ mod tests {
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_ok());
         assert!(XPathBuf::from(format!("/proc/{this}/task"))
-            .check(this, Some(&FileType::Dir), Some(&xpath!("{this}")), true,)
+            .check(
+                this,
+                Some(&FileType::Dir),
+                Some(&xpath!("{this}")),
+                true,
+                true
+            )
             .is_ok());
         assert!(XPathBuf::from("/proc")
-            .check(this, Some(&FileType::Dir), Some(&xpath!("{that}")), true,)
+            .check(
+                this,
+                Some(&FileType::Dir),
+                Some(&xpath!("{that}")),
+                true,
+                true
+            )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{that}"))
             .check(
@@ -1570,6 +1727,7 @@ mod tests {
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_ok());
         assert!(XPathBuf::from(format!("/proc/{that}"))
@@ -1578,10 +1736,17 @@ mod tests {
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_ok());
         assert!(XPathBuf::from(format!("/proc/{that}/task"))
-            .check(this, Some(&FileType::Dir), Some(&xpath!("{that}")), true,)
+            .check(
+                this,
+                Some(&FileType::Dir),
+                Some(&xpath!("{that}")),
+                true,
+                true
+            )
             .is_ok());
     }
 
@@ -1589,7 +1754,13 @@ mod tests {
     fn test_path_check_procfs_syd_leader() {
         let syd = Pid::this();
         assert!(XPathBuf::from("/proc")
-            .check(syd, Some(&FileType::Dir), Some(&xpath!("{syd}")), true,)
+            .check(
+                syd,
+                Some(&FileType::Dir),
+                Some(&xpath!("{syd}")),
+                true,
+                true
+            )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{syd}"))
             .check(
@@ -1597,6 +1768,7 @@ mod tests {
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{syd}"))
@@ -1605,10 +1777,17 @@ mod tests {
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{syd}/task"))
-            .check(syd, Some(&FileType::Dir), Some(&xpath!("{syd}")), true,)
+            .check(
+                syd,
+                Some(&FileType::Dir),
+                Some(&xpath!("{syd}")),
+                true,
+                true
+            )
             .is_err());
     }
 
@@ -1624,7 +1803,13 @@ mod tests {
             rx.recv().unwrap()
         };
         assert!(XPathBuf::from("/proc")
-            .check(tid, Some(&FileType::Dir), Some(&xpath!("{tid}")), true,)
+            .check(
+                tid,
+                Some(&FileType::Dir),
+                Some(&xpath!("{tid}")),
+                true,
+                true
+            )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{tid}"))
             .check(
@@ -1632,6 +1817,7 @@ mod tests {
                 Some(&FileType::Reg),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{tid}"))
@@ -1640,10 +1826,17 @@ mod tests {
                 Some(&FileType::Dir),
                 Some(&XPath::from_bytes(b"")),
                 true,
+                true,
             )
             .is_err());
         assert!(XPathBuf::from(format!("/proc/{tid}/task"))
-            .check(tid, Some(&FileType::Dir), Some(&xpath!("{tid}")), true,)
+            .check(
+                tid,
+                Some(&FileType::Dir),
+                Some(&xpath!("{tid}")),
+                true,
+                true
+            )
             .is_err());
     }
 
@@ -1808,6 +2001,7 @@ mod tests {
             "valid_filename.txt",
             "hello_world",
             "File123",
+            "Makefile",
             "こんにちは", // Japanese characters
             "文件",       // Chinese characters
             "emoji😀",    // Starts with permitted character
@@ -1819,7 +2013,6 @@ mod tests {
             "test_file",
             "file name",
             "file☃name",    // Snowman character
-            "\u{00A0}",     // Non-breaking space
             "name\u{0080}", // Contains 0x80 (allowed)
             "name\u{00FE}", // Contains 0xFE (allowed)
             "😀name",       // Multi-byte character at start
@@ -1830,6 +2023,12 @@ mod tests {
             "name-",        // Hyphen at end (allowed)
             "name_",        // Underscore at end (allowed)
             "name.",        // Period at end (allowed)
+            "a\u{0020}b",   // SPACE in the middle (allowed)
+            "a\u{00A0}b",   // NO-BREAK SPACE in the middle (allowed)
+            "a\u{1680}b",   // OGHAM SPACE MARK in the middle (allowed)
+            "a\u{2007}b",   // FIGURE SPACE in the middle (allowed)
+            "a\u{202F}b",   // NARROW NO-BREAK SPACE in the middle (allowed)
+            "a\u{3000}b",   // IDEOGRAPHIC SPACE in the middle (allowed)
         ];
 
         for (idx, name) in valid_filenames.iter().enumerate() {
@@ -1844,73 +2043,80 @@ mod tests {
     #[test]
     fn test_check_name_invalid() {
         let invalid_filenames: &[&[u8]] = &[
-            b"",                 // Empty filename
-            b"-",                // Starts with '-'
-            b"*",                // Starts with '*'
-            b"?",                // Starts with '?'
-            b"!",                // Starts with '!'
-            b"$",                // Starts with '$'
-            b"`",                // Starts with '`'
-            b" -",               // Starts with space
-            b"~home",            // Starts with '~'
-            b"*home",            // Starts with '*'
-            b"?home",            // Starts with '?'
-            b"!home",            // Starts with '!'
-            b"$home",            // Starts with '$'
-            b"`home",            // Starts with '`'
-            b"file ",            // Ends with space
-            b"file*",            // Ends with '*'
-            b"file?",            // Ends with '?'
-            b"file!",            // Ends with '!'
-            b"file$",            // Ends with '$'
-            b"file`",            // Ends with '`'
-            b"bad*name",         // Contains '*'
-            b"bad?name",         // Contains '?'
-            b"bad!name",         // Contains '!'
-            b"bad$name",         // Contains '$'
-            b"bad`name",         // Contains '`'
-            b"bad\nname",        // Contains newline
-            b"\0",               // Null byte
-            b"bad\0name",        // Contains null byte
-            b"bad\x7Fname",      // Contains delete character
-            b"bad\xFFname",      // Contains 0xFF
-            b"\x1Fcontrol",      // Starts with control character
-            b"name\x1F",         // Ends with control character
-            b"name\x7F",         // Ends with delete character
-            b"name\xFF",         // Ends with 0xFF
-            b"name ",            // Ends with space
-            b"-name",            // Starts with '-'
-            b" name",            // Starts with space
-            b"~name",            // Starts with '~'
-            b"*name",            // Starts with '*'
-            b"?name",            // Starts with '?'
-            b"!name",            // Starts with '!'
-            b"$name",            // Starts with '$'
-            b"`name",            // Starts with '`'
-            b"name\x19",         // Contains control character
-            b"name\n",           // Ends with newline
-            b"\nname",           // Starts with newline
-            b"na\nme",           // Contains newline
-            b"name\t",           // Contains tab
-            b"name\r",           // Contains carriage return
-            b"name\x1B",         // Contains escape character
-            b"name\x00",         // Contains null byte
-            b"name\x7F",         // Contains delete character
-            b"name\xFF",         // Contains 0xFF (disallowed)
-            b"\xFF",             // Single byte 0xFF
-            b"name\x80\xFF",     // Contains valid and invalid extended ASCII
-            b"name\xC0\xAF",     // Invalid UTF-8 sequence
-            b"\xF0\x28\x8C\xBC", // Invalid UTF-8 sequence
-            b"\xF0\x90\x28\xBC", // Invalid UTF-8 sequence
-            b"\xF0\x28\x8C\x28", // Invalid UTF-8 sequence
-            b"name\xFFname",     // Contains 0xFF
-            b"name\xC3\x28",     // Invalid UTF-8 sequence
-            b"name\xA0\xA1",     // Invalid UTF-8 sequence
-            b"\xE2\x28\xA1",     // Invalid UTF-8 sequence
-            b"\xE2\x82\x28",     // Invalid UTF-8 sequence
-            b"\xF0\x28\x8C\xBC", // Invalid UTF-8 sequence
-            b"\xF0\x90\x28\xBC", // Invalid UTF-8 sequence
-            b"\xF0\x28\x8C\x28", // Invalid UTF-8 sequence
+            b"",                             // Empty filename
+            b"-",                            // Starts with '-'
+            b"*",                            // Starts with '*'
+            b"?",                            // Starts with '?'
+            b"!",                            // Starts with '!'
+            b"$",                            // Starts with '$'
+            b"`",                            // Starts with '`'
+            b" -",                           // Starts with space
+            b"~home",                        // Starts with '~'
+            b"*home",                        // Starts with '*'
+            b"?home",                        // Starts with '?'
+            b"!home",                        // Starts with '!'
+            b"$home",                        // Starts with '$'
+            b"`home",                        // Starts with '`'
+            b"file ",                        // Ends with space
+            b"file*",                        // Ends with '*'
+            b"file?",                        // Ends with '?'
+            b"file!",                        // Ends with '!'
+            b"file$",                        // Ends with '$'
+            b"file`",                        // Ends with '`'
+            b"bad*name",                     // Contains '*'
+            b"bad?name",                     // Contains '?'
+            b"bad!name",                     // Contains '!'
+            b"bad$name",                     // Contains '$'
+            b"bad`name",                     // Contains '`'
+            b"bad\nname",                    // Contains newline
+            b"\0",                           // Null byte
+            b"bad\0name",                    // Contains null byte
+            b"bad\x7Fname",                  // Contains delete character
+            b"bad\xFFname",                  // Contains 0xFF
+            b"\x1Fcontrol",                  // Starts with control character
+            b"name\x1F",                     // Ends with control character
+            b"name\x7F",                     // Ends with delete character
+            b"name\xFF",                     // Ends with 0xFF
+            b"name ",                        // Ends with space
+            b"-name",                        // Starts with '-'
+            b" name",                        // Starts with space
+            b"~name",                        // Starts with '~'
+            b"*name",                        // Starts with '*'
+            b"?name",                        // Starts with '?'
+            b"!name",                        // Starts with '!'
+            b"$name",                        // Starts with '$'
+            b"`name",                        // Starts with '`'
+            b"name\x19",                     // Contains control character
+            b"name\n",                       // Ends with newline
+            b"\nname",                       // Starts with newline
+            b"na\nme",                       // Contains newline
+            b"name\t",                       // Contains tab
+            b"name\r",                       // Contains carriage return
+            b"name\x1B",                     // Contains escape character
+            b"name\x00",                     // Contains null byte
+            b"name\x7F",                     // Contains delete character
+            b"name\xFF",                     // Contains 0xFF (disallowed)
+            b"\xFF",                         // Single byte 0xFF
+            b"name\x80\xFF",                 // Contains valid and invalid extended ASCII
+            b"name\xC0\xAF",                 // Invalid UTF-8 sequence
+            b"\xF0\x28\x8C\xBC",             // Invalid UTF-8 sequence
+            b"\xF0\x90\x28\xBC",             // Invalid UTF-8 sequence
+            b"\xF0\x28\x8C\x28",             // Invalid UTF-8 sequence
+            b"name\xFFname",                 // Contains 0xFF
+            b"name\xC3\x28",                 // Invalid UTF-8 sequence
+            b"name\xA0\xA1",                 // Invalid UTF-8 sequence
+            b"\xE2\x28\xA1",                 // Invalid UTF-8 sequence
+            b"\xE2\x82\x28",                 // Invalid UTF-8 sequence
+            b"\xF0\x28\x8C\xBC",             // Invalid UTF-8 sequence
+            b"\xF0\x90\x28\xBC",             // Invalid UTF-8 sequence
+            b"\xF0\x28\x8C\x28",             // Invalid UTF-8 sequence
+            b"\xC2\xA0",                     // Non-breaking space
+            b"\x20file",                     // leading SPACE U+0020
+            b"file\x20",                     // trailing SPACE U+0020
+            b"\xC2\xA0file",                 // leading NO-BREAK SPACE U+00A0
+            b"file\xE3\x80\x80",             // trailing IDEOGRAPHIC SPACE U+3000
+            b"\xE2\x80\xAFfile",             // leading NARROW NO-BREAK SPACE U+202F
+            b"\xE2\x81\x9Ffile\xE2\x81\x9F", // both sides MEDIUM MATHEMATICAL SPACE U+205F
         ];
 
         for (idx, name) in invalid_filenames.iter().enumerate() {
@@ -1963,8 +2169,7 @@ mod tests {
     fn test_check_name_edge_cases() {
         // Filenames with length 1
         let valid_single_chars = [
-            "a", "b", "Z", "9", "_", ".", "😀",       // Valid multi-byte character
-            "\u{00A0}", // Non-breaking space
+            "a", "b", "Z", "9", "_", ".", "😀", // Valid multi-byte character
         ];
 
         for (idx, name) in valid_single_chars.iter().enumerate() {
@@ -1976,17 +2181,24 @@ mod tests {
         }
 
         let invalid_single_chars: &[&[u8]] = &[
-            b"-",    // Starts with '-'
-            b" ",    // Space character
-            b"~",    // Tilde character
-            b"*",    // Starts with '*'
-            b"?",    // Starts with '?'
-            b"\n",   // Newline character
-            b"\r",   // Newline character
-            b"\x7F", // Delete character
-            b"\x1F", // Control character
-            b"\xFF", // 0xFF disallowed
-            b"\0",   // Null byte
+            b"-",            // Starts with '-'
+            b" ",            // Space character
+            b"~",            // Tilde character
+            b"*",            // Starts with '*'
+            b"?",            // Starts with '?'
+            b"\n",           // Newline character
+            b"\r",           // Newline character
+            b"\x7F",         // Delete character
+            b"\x1F",         // Control character
+            b"\xFF",         // 0xFF disallowed
+            b"\0",           // Null byte
+            b"\xC2\xA0",     // Non-breaking space
+            b"\x20",         // SPACE U+0020
+            b"\xC2\xA0",     // NO-BREAK SPACE U+00A0
+            b"\xE1\x9A\x80", // OGHAM SPACE MARK U+1680
+            b"\xE2\x80\x87", // FIGURE SPACE U+2007
+            b"\xE2\x80\xAF", // NARROW NO-BREAK SPACE U+202F
+            b"\xE3\x80\x80", // IDEOGRAPHIC SPACE U+3000
         ];
 
         for (idx, name) in invalid_single_chars.iter().enumerate() {
@@ -1997,4 +2209,132 @@ mod tests {
             );
         }
     }
+
+    // Helper to build a buffer.
+    fn xp(bytes: &[u8]) -> XPathBuf {
+        XPathBuf(bytes.to_vec())
+    }
+
+    #[test]
+    fn test_replace_prefix_no_match_is_noop() -> Result<(), Errno> {
+        let mut x = xp(b"/ordinary/path");
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        assert_eq!(x.as_bytes(), b"/ordinary/path");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_equal_length_overwrite() -> Result<(), Errno> {
+        let mut x = xp(b"abc:/tail");
+        x.replace_prefix(b"abc:", b"XYZ:")?;
+        assert_eq!(x.as_bytes(), b"XYZ:/tail");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_grow_old_to_new() -> Result<(), Errno> {
+        let mut x = xp(b"!memfd:/some/path");
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        assert_eq!(x.as_bytes(), b"!memfd-hugetlb:/some/path");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_grow_when_only_prefix_present() -> Result<(), Errno> {
+        let mut x = xp(b"!memfd:");
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        assert_eq!(x.as_bytes(), b"!memfd-hugetlb:");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_idempotent_for_upgrade() -> Result<(), Errno> {
+        // Already upgraded; should not duplicate "-hugetlb".
+        let mut x = xp(b"!memfd-hugetlb:/already");
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        assert_eq!(x.as_bytes(), b"!memfd-hugetlb:/already");
+        // Re-apply again to ensure stability.
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        assert_eq!(x.as_bytes(), b"!memfd-hugetlb:/already");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_shrink_old_to_new() -> Result<(), Errno> {
+        let mut x = xp(b"prefix-long:/rest");
+        x.replace_prefix(b"prefix-long:", b"p:")?;
+        assert_eq!(x.as_bytes(), b"p:/rest");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_shrink_when_new_is_prefix_of_old() -> Result<(), Errno> {
+        // Ensure we still shrink even though the buffer "starts_with(new)".
+        let mut x = xp(b"foobar-long:/zzz");
+        x.replace_prefix(b"foobar-long:", b"foobar:")?;
+        assert_eq!(x.as_bytes(), b"foobar:/zzz");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_tail_integrity_grow() -> Result<(), Errno> {
+        let mut x = xp(b"!memfd:/a/b/c/d/e");
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        assert_eq!(x.as_bytes(), b"!memfd-hugetlb:/a/b/c/d/e");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_tail_integrity_shrink() -> Result<(), Errno> {
+        let mut x = xp(b"longerprefix:/a/b/c");
+        x.replace_prefix(b"longerprefix:", b"lp:")?;
+        assert_eq!(x.as_bytes(), b"lp:/a/b/c");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_capacity_growth_handled() -> Result<(), Errno> {
+        // Force the "grow" code path to hit reserve/resize.
+        let mut x = XPathBuf::with_capacity(1);
+        x.append_bytes(b"!memfd:/x/y/z");
+        let cap_before = x.capacity();
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        let cap_after = x.capacity();
+        assert!(cap_after >= cap_before);
+        assert_eq!(x.as_bytes(), b"!memfd-hugetlb:/x/y/z");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_large_tail_stress() -> Result<(), Errno> {
+        // Build a long tail to stress copy_within correctness.
+        let mut tail = Vec::new();
+        for _ in 0..512 {
+            tail.extend_from_slice(b"/component");
+        }
+        let mut p = b"!memfd:".to_vec();
+        p.extend_from_slice(&tail);
+        let mut x = xp(&p);
+        x.replace_prefix(b"!memfd:", b"!memfd-hugetlb:")?;
+        let mut expected = b"!memfd-hugetlb:".to_vec();
+        expected.extend_from_slice(&tail);
+        assert_eq!(x.as_bytes(), expected.as_slice());
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_utf8_agnostic_bytes() -> Result<(), Errno> {
+        // Ensure function doesn't assume UTF-8; operate on arbitrary bytes.
+        let mut x = xp(b"\xF0\x9F\x92\xA9prefix-long:\xFF\x00tail");
+        x.replace_prefix(b"\xF0\x9F\x92\xA9prefix-long:", b"\xF0\x9F\x92\xA9p:")?;
+        assert_eq!(x.as_bytes(), b"\xF0\x9F\x92\xA9p:\xFF\x00tail");
+        Ok(())
+    }
+
+    #[test]
+    fn test_replace_prefix_empty_old_is_error() {
+        let mut x = xp(b"!memfd:/whatever");
+        let res = x.replace_prefix(b"", b"!memfd-hugetlb:");
+        assert!(matches!(res, Err(Errno::EINVAL)));
+    }
 }
