@@ -14,13 +14,12 @@
 // Commit:d56805869ba3cbe47021d5660bbaf19ac5ec4bfb
 
 use std::{
-    collections::HashMap,
     env,
     fs::OpenOptions,
     io::Write,
     option::Option,
     os::{
-        fd::{AsRawFd, BorrowedFd, RawFd},
+        fd::{AsRawFd, RawFd},
         unix::fs::OpenOptionsExt,
     },
     sync::{
@@ -30,25 +29,23 @@ use std::{
     thread,
 };
 
+use libseccomp::ScmpFilterContext;
 use nix::{
     errno::Errno,
     fcntl::OFlag,
     sched::{unshare, CloneFlags},
-    sys::{
-        epoll::Epoll,
-        signal::{sigaction, SaFlags, SigAction, SigHandler, SigSet, Signal},
-    },
-    unistd::{getpid, pipe2, read},
+    sys::signal::{sigaction, SaFlags, SigAction, SigHandler, SigSet, Signal},
+    unistd::{getpid, pipe2, read, Gid, Uid},
 };
 
 use crate::{
     alert,
-    compat::epoll_ctl_safe,
     config::*,
-    crit,
-    err::{err2no, SydJoinHandle, SydResult},
+    confine::ExportMode,
+    err::{err2no, scmp2no, SydJoinHandle, SydResult},
     error,
     fs::{block_signal, closeexcept, retry_on_eintr, retry_on_intr, seccomp_export_pfc},
+    hash::SydHashMap,
     hook::HandlerMap,
     info, rwrite, rwriteln,
     sandbox::{Capability, Flags, Sandbox},
@@ -57,27 +54,8 @@ use crate::{
         emu::Worker,
         int::Interrupter,
         ipc::IpcWorker,
-        BindMap, WorkerCache, WorkerData,
+        UnixMap, WorkerCache, WorkerData,
     },
-    ExportMode,
-};
-
-// Epoll event to add seccomp fd to epoll (becomes readable when system
-// call is interrupted). We specifically zero out the data field to
-// distinguish from PidFds.
-//
-// Quoting: https://idea.popcount.org/2017-02-20-epoll-is-fundamentally-broken-12/
-// """
-// The best and the only scalable approach is to use recent
-// Kernel 4.5+ and use level-triggered events with
-// EPOLLEXCLUSIVE flag. This will ensure only one thread is
-// woken for an event, avoid "thundering herd" issue and scale
-// properly across multiple CPU's.
-// """/
-#[allow(clippy::cast_sign_loss)]
-const SCMP_EPOLL_EVENT: libc::epoll_event = libc::epoll_event {
-    events: (libc::EPOLLIN | libc::EPOLLEXCLUSIVE) as u32,
-    u64: 0, // zeroed out to distinguish from pid-fds.
 };
 
 // Signal handler function for SIGALRM.
@@ -90,21 +68,18 @@ pub(crate) struct ThreadPool {
     keep_alive: u16,
     flags: Flags,
     seccomp_fd: RawFd,
-    pub(crate) epoll: Arc<Epoll>,
     pub(crate) cache: Arc<WorkerCache<'static>>,
     sandbox: Arc<RwLock<Sandbox>>,
     handlers: Arc<HandlerMap>,
-    bind_map: BindMap,
     crypt_map: Option<AesMap>,
+    unix_map: UnixMap,
     should_exit: Arc<AtomicBool>,
     worker_data: Arc<WorkerData>,
 }
 
 impl ThreadPool {
     /// Construct a new `ThreadPool` with the specified core pool size,
-    /// max pool size and keep_alive time for non-core threads. This
-    /// function creates an epoll instance and adds the seccomp fd to it
-    /// but it does not spawn any threads.
+    /// max pool size and keep_alive time for non-core threads.
     ///
     /// `core_size` specifies the amount of threads to keep alive for as
     /// long as the `ThreadPool` exists and the seccomp fd remains open.
@@ -115,7 +90,6 @@ impl ThreadPool {
     #[allow(clippy::cognitive_complexity)]
     #[allow(clippy::too_many_arguments)]
     pub(crate) fn new(
-        epoll: Epoll,
         seccomp_fd: RawFd,
         flags: Flags,
         core_size: usize,
@@ -124,28 +98,20 @@ impl ThreadPool {
         handlers: Arc<HandlerMap>,
         should_exit: Arc<AtomicBool>,
         crypt_map: Option<AesMap>,
-    ) -> Result<Self, Errno> {
-        // SAFETY: Borrow FD to make I/O safe API hippie.
-        let seccomp_fd = unsafe { BorrowedFd::borrow_raw(seccomp_fd) };
-        epoll_ctl_safe(&epoll.0, seccomp_fd.as_raw_fd(), Some(SCMP_EPOLL_EVENT))?;
-
-        let epoll = Arc::new(epoll);
-        let cache = Arc::new(WorkerCache::new());
-
-        Ok(Self {
-            cache,
+    ) -> Self {
+        Self {
             sandbox,
             crypt_map,
             handlers,
             core_size,
             keep_alive,
-            epoll,
             flags,
             should_exit,
-            seccomp_fd: seccomp_fd.as_raw_fd(),
+            seccomp_fd,
+            cache: Arc::new(WorkerCache::new()),
             worker_data: Arc::new(WorkerData::default()),
-            bind_map: Arc::new(RwLock::new(HashMap::default())),
-        })
+            unix_map: Arc::new(RwLock::new(SydHashMap::default())),
+        }
     }
 
     /// Boot the thread pool. This is the main entry point.
@@ -159,12 +125,7 @@ impl ThreadPool {
             Some(ExportMode::BerkeleyPacketFilter) => {
                 // Worker rules
                 let is_crypt = self.crypt_map.is_some();
-                let ctx = Worker::prepare_confine(
-                    self.seccomp_fd,
-                    self.epoll.0.as_raw_fd(),
-                    self.flags,
-                    is_crypt,
-                )?;
+                let ctx = Worker::prepare_confine(self.seccomp_fd, self.flags, is_crypt, &[], &[])?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -174,8 +135,14 @@ impl ThreadPool {
 
                 // Interrupter rules
                 // We pass dry_run=true to avoid Landlock confinement.
-                let ctx =
-                    Interrupter::prepare_confine(self.seccomp_fd, getpid(), self.flags, true)?;
+                let ctx = Interrupter::prepare_confine(
+                    self.seccomp_fd,
+                    getpid(),
+                    self.flags,
+                    &[],
+                    &[],
+                    true,
+                )?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -186,7 +153,7 @@ impl ThreadPool {
                 // IPC thread rules
                 // We pass dummy RawFd=2525 for epoll FD.
                 // We pass dry_run=true to avoid Landlock confinement.
-                let ctx = IpcWorker::prepare_confine(2525, self.flags, true)?;
+                let ctx = IpcWorker::prepare_confine(2525, self.flags, &[], &[], true)?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -195,7 +162,7 @@ impl ThreadPool {
                 ctx.export_bpf(file)?;
 
                 // Aes worker rules
-                let ctx = AesWorker::prepare_confine(self.flags)?;
+                let ctx = AesWorker::prepare_confine(self.flags, &[], &[])?;
                 let file = OpenOptions::new()
                     .write(true)
                     .create_new(true)
@@ -213,12 +180,7 @@ impl ThreadPool {
                     self.seccomp_fd
                 )?;
                 let is_crypt = self.crypt_map.is_some();
-                let ctx = Worker::prepare_confine(
-                    self.seccomp_fd,
-                    self.epoll.0.as_raw_fd(),
-                    self.flags,
-                    is_crypt,
-                )?;
+                let ctx = Worker::prepare_confine(self.seccomp_fd, self.flags, is_crypt, &[], &[])?;
                 rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
                 // We pass dry_run=true to avoid Landlock confinement.
@@ -227,18 +189,24 @@ impl ThreadPool {
                     "# Syd interrupter rules with seccomp fd {}",
                     self.seccomp_fd
                 )?;
-                let ctx =
-                    Interrupter::prepare_confine(self.seccomp_fd, getpid(), self.flags, true)?;
+                let ctx = Interrupter::prepare_confine(
+                    self.seccomp_fd,
+                    getpid(),
+                    self.flags,
+                    &[],
+                    &[],
+                    true,
+                )?;
                 rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
                 // We pass dummy RawFd=2525 for epoll FD.
                 // We pass dry_run=true to avoid Landlock confinement.
                 rwriteln!(stdout, "# Syd ipc rules")?;
-                let ctx = IpcWorker::prepare_confine(2525, self.flags, true)?;
+                let ctx = IpcWorker::prepare_confine(2525, self.flags, &[], &[], true)?;
                 rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
 
                 rwriteln!(stdout, "# Syd encryptor rules")?;
-                let ctx = AesWorker::prepare_confine(self.flags)?;
+                let ctx = AesWorker::prepare_confine(self.flags, &[], &[])?;
                 rwrite!(stdout, "{}", seccomp_export_pfc(&ctx)?)?;
             }
             _ => {}
@@ -272,31 +240,20 @@ impl ThreadPool {
             .name("syd_mon".to_string())
             .stack_size(MON_STACK_SIZE)
             .spawn(move || {
-                // SAFETY: Default panic hook wont work with stdio(3) closed.
-                std::panic::set_hook(Box::new(|info| {
-                    let err = match info.payload().downcast_ref::<&'static str>() {
-                        Some(s) => *s,
-                        None => match info.payload().downcast_ref::<String>() {
-                            Some(s) => &**s,
-                            None => "?",
-                        },
+                // Get information from the sandbox about:
+                // 1. Encryption if Crypt sandboxing is enabled.
+                // 2. Transit UIDs and GIDs for SafeSetId.
+                let (fpid, crypt, transit_uids, transit_gids) = {
+                    let sandbox = self.sandbox.read().unwrap_or_else(|err| err.into_inner());
+                    let fpid = sandbox.fpid;
+
+                    let crypt = if sandbox.enabled(Capability::CAP_CRYPT) {
+                        Some((sandbox.crypt_setup()?, sandbox.crypt_tmp))
+                    } else {
+                        None
                     };
 
-                    let file = info.location().map(|l| l.file());
-                    let line = info.location().map(|l| l.line());
-
-                    crit!("ctx": "panic", "op": "panic_monitor_thread",
-                        "msg": err, "file": file, "line": line);
-                }));
-
-                // Get information on encryption from the sandbox.
-                let (crypt, fpid) = {
-                    let sandbox = self.sandbox.read().unwrap_or_else(|err| err.into_inner());
-                    if sandbox.enabled(Capability::CAP_CRYPT) {
-                        (Some((sandbox.crypt_setup()?, sandbox.crypt_tmp)), sandbox.fpid)
-                    } else {
-                        (None, sandbox.fpid)
-                    }
+                    (fpid, crypt, sandbox.transit_uids.clone(), sandbox.transit_gids.clone())
                 };
 
                 // SAFETY: We use exit_group(2) here to bail,
@@ -317,7 +274,6 @@ impl ThreadPool {
                     NULL_FD() as libc::c_uint,
                     fpid as libc::c_uint,
                     self.seccomp_fd as libc::c_uint,
-                    self.epoll.0.as_raw_fd() as libc::c_uint,
                     crate::log::LOG_FD.load(Ordering::Relaxed) as libc::c_uint,
                 ];
                 #[allow(clippy::cast_sign_loss)]
@@ -330,15 +286,16 @@ impl ThreadPool {
                 }
                 set.sort_unstable();
                 closeexcept(&set)?;
+                drop(set);
 
                 // Spawn the interrupt thread which will confine itself.
-                self.try_spawn_interrupt()?;
+                self.try_spawn_interrupt(&transit_uids, &transit_gids)?;
 
                 // Spawn the AES thread if encryption is on.
                 let crypt_handle = if let Some((fds, tmp)) = crypt {
                     let map = self.crypt_map.as_ref().map(Arc::clone).ok_or(Errno::ENOKEY)?;
                     let should_exit = Arc::clone(&self.should_exit);
-                    Some(self.try_spawn_aes(fds, map, tmp.is_none(), should_exit)?)
+                    Some(self.try_spawn_aes(fds, map, tmp.is_none(), should_exit, &transit_uids, &transit_gids)?)
                 } else {
                     None
                 };
@@ -350,30 +307,64 @@ impl ThreadPool {
                     "core_size": self.core_size,
                     "keep_alive": self.keep_alive);
 
-                // SAFETY: If sandbox is locked, confine right away.
-                // Pass confined parameter to try_spawn so subsequent
-                // spawned threads don't need to reapply the same filter
-                // as it is inherited.
+                // SAFETY:
+                // 1. If sandbox is locked, confine right away.
+                //    Pass confined parameter to try_spawn so subsequent
+                //    spawned threads don't need to reapply the same filter
+                //    as it is inherited.
+                // 2. If sandbox is not locked yet, build the seccomp context anyway,
+                //    precompute it and pass it to emulator threads for fast confinement.
                 let dry_run = env::var_os(ENV_SKIP_SCMP).is_some() || ExportMode::from_env().is_some();
-                let is_crypt = self.crypt_map.is_some();
                 let safe_setid = self.flags.intersects(Flags::FL_ALLOW_SAFE_SETUID | Flags::FL_ALLOW_SAFE_SETGID);
-                let mut confined = if Sandbox::locked_once() {
-                    let ctx = Worker::prepare_confine(self.seccomp_fd, self.epoll.0.as_raw_fd(), self.flags, is_crypt)?;
-                    if !dry_run {
+                let is_crypt = self.crypt_map.is_some();
+
+                let mut ctx = if !dry_run {
+                    let ctx = Worker::prepare_confine(self.seccomp_fd, self.flags, is_crypt, &transit_uids, &transit_gids)?;
+
+                    if Sandbox::locked_once() {
+                        // Sandbox locked, confine right away.
+                        //
+                        // SAFETY: We use exit_group(2) here to bail,
+                        // because this confinement is a critical safety feature.
+                        if let Err(error) = ctx.load() {
+                            let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
+                            alert!("ctx": "boot", "op": "confine_monitor_thread",
+                                "msg": format!("failed to confine: {error}"),
+                                "err": errno as i32);
+                            std::process::exit(101);
+                        }
+
                         info!("ctx": "confine", "op": "confine_monitor_thread",
                             "msg": format!("monitor thread confined with{} SROP mitigation",
                                 if safe_setid { "out" } else { "" }));
-                        ctx.load()?;
+
+                        None
                     } else {
-                        error!("ctx": "confine", "op": "confine_monitor_thread",
-                            "msg": "monitor thread is running unconfined in debug mode");
+                        // Sandbox not locked yet, precompute and save filter.
+                        //
+                        // SAFETY: We use exit_group(2) here to bail,
+                        // because this confinement is a critical safety feature.
+                        #[cfg(libseccomp_v2_6)]
+                        if let Err(error) = ctx.precompute() {
+                            let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
+                            alert!("ctx": "boot", "op": "confine_monitor_thread",
+                                "msg": format!("failed to precompute: {error}"),
+                                "err": errno as i32);
+                            std::process::exit(101);
+                        }
+
+                        info!("ctx": "confine", "op": "confine_monitor_thread",
+                            "msg": "monitor thread is running unconfined because sandbox isn't locked yet");
+
+                        Some(ctx)
                     }
-                    true
                 } else {
-                    info!("ctx": "confine", "op": "confine_monitor_thread",
-                        "msg": "monitor thread is running unconfined because sandbox isn't locked yet");
-                    false
+                    error!("ctx": "confine", "op": "confine_monitor_thread",
+                        "msg": "monitor thread is running unconfined in debug mode");
+                    None
                 };
+                drop(transit_uids);
+                drop(transit_gids);
 
                 info!("ctx": "boot", "op": "start_core_emulator_threads",
                     "msg": format!("starting {} core emulator thread{}, sandboxing started!",
@@ -383,27 +374,37 @@ impl ThreadPool {
                     "keep_alive": self.keep_alive);
 
                 // Spawn the initial core emulator thread.
-                self.try_spawn(confined).map(drop)?;
+                self.try_spawn(ctx.as_ref()).map(drop)?;
 
                 // Wait for grace period to give the initial
                 // core emulator thread a chance to spawn itself.
                 std::thread::sleep(MON_GRACE_TIME);
 
                 loop {
-                    // Confine as necessary.
-                    #[allow(clippy::disallowed_methods)]
-                    if !confined && Sandbox::locked_once() {
-                        let ctx = Worker::prepare_confine(self.seccomp_fd, self.epoll.0.as_raw_fd(), self.flags, is_crypt)?;
-                        if !dry_run {
-                            ctx.load()?;
+                    // Confine and drop filter if sandbox is locked.
+                    if let Some(ref filter) = ctx {
+                        if Sandbox::locked_once() {
+                            // SAFETY: We use exit_group(2) here to bail,
+                            // because this confinement is a critical safety feature.
+                            if let Err(error) = filter.load() {
+                                let errno = scmp2no(&error).unwrap_or(Errno::ENOSYS);
+                                alert!("ctx": "boot", "op": "confine_monitor_thread",
+                                    "msg": format!("failed to confine: {error}"),
+                                    "err": errno as i32);
+                                std::process::exit(101);
+                            }
+
                             info!("ctx": "confine", "op": "confine_monitor_thread",
                                 "msg": format!("monitor thread confined with{} SROP mitigation",
                                     if safe_setid { "out" } else { "" }));
-                        } else {
-                            error!("ctx": "confine", "op": "confine_monitor_thread",
-                                "msg": "monitor thread is running unconfined in debug mode");
+
+                            // SAFETY: We cannot free the seccomp context here,
+                            // because it may have references in emulator
+                            // threads.
+                            std::mem::forget(ctx);
+
+                            ctx = None;
                         }
-                        confined = true;
                     }
 
                     if self.should_exit.load(Ordering::Relaxed) {
@@ -412,7 +413,7 @@ impl ThreadPool {
                     }
 
                     // Spawn a new thread if all others are busy.
-                    match self.try_spawn(confined) {
+                    match self.try_spawn(ctx.as_ref()) {
                         Ok(Some(_)) => {
                             // We have spawned a new emulator thread,
                             // wait for one cycle before reattempting.
@@ -444,7 +445,11 @@ impl ThreadPool {
     /// Spawn an interrupt handler thread to unblock Syd syscall
     /// handler threads when the respective sandbox process
     /// receives a non-restarting signal.
-    pub(crate) fn try_spawn_interrupt(&self) -> SydResult<SydJoinHandle<()>> {
+    pub(crate) fn try_spawn_interrupt(
+        &self,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
+    ) -> SydResult<SydJoinHandle<()>> {
         // Block SIGALRM, this mask will be inherited by emulators.
         block_signal(Signal::SIGALRM)?;
 
@@ -467,6 +472,8 @@ impl ThreadPool {
             Interrupter::new(
                 self.seccomp_fd,
                 self.flags,
+                transit_uids,
+                transit_gids,
                 Arc::clone(&self.should_exit),
                 Arc::clone(&self.cache),
             )
@@ -498,8 +505,18 @@ impl ThreadPool {
         files: AesMap,
         memfd: bool,
         should_exit: Arc<AtomicBool>,
+        transit_uids: &[(Uid, Uid)],
+        transit_gids: &[(Gid, Gid)],
     ) -> Result<SydJoinHandle<()>, Errno> {
-        let worker = AesWorker::new(fdalg, files, self.flags, memfd, should_exit);
+        let worker = AesWorker::new(
+            fdalg,
+            files,
+            self.flags,
+            memfd,
+            should_exit,
+            transit_uids,
+            transit_gids,
+        );
 
         // AesWorker has only RawFds as Fds which
         // we do _not_ want to duplicate on clone,
@@ -512,7 +529,10 @@ impl ThreadPool {
     /// The boolean in the success case is true if the thread we spawned was a core thread.
     #[allow(clippy::cognitive_complexity)]
     #[allow(clippy::type_complexity)]
-    pub(crate) fn try_spawn(&self, confined: bool) -> SydResult<Option<(SydJoinHandle<()>, bool)>> {
+    pub(crate) fn try_spawn(
+        &self,
+        ctx: Option<&ScmpFilterContext>,
+    ) -> SydResult<Option<(SydJoinHandle<()>, bool)>> {
         // Create a new worker if there are no idle threads and the
         // current worker count is lower than the max pool size.
         let worker_count_val = self.worker_data.0.load(Ordering::Relaxed);
@@ -560,17 +580,16 @@ impl ThreadPool {
             retry_on_intr(|| {
                 Worker::new(
                     self.seccomp_fd,
-                    Arc::clone(&self.epoll),
                     Arc::clone(&self.cache),
                     Arc::clone(&self.sandbox),
                     Arc::clone(&self.handlers),
                     keep_alive,
                     Arc::clone(&self.should_exit),
                     Arc::clone(&self.worker_data),
-                    Arc::clone(&self.bind_map),
                     self.crypt_map.as_ref().map(Arc::clone),
+                    Arc::clone(&self.unix_map),
                 )
-                .try_spawn(confined)
+                .try_spawn(ctx)
             })?,
             keep_alive.is_none(),
         )))
