@@ -26,6 +26,7 @@
 
 use std::{
     ffi::{CStr, OsStr, OsString},
+    fmt,
     fs::File,
     io::{BufReader, IoSlice, IoSliceMut, Read, Seek, SeekFrom},
     num::NonZeroUsize,
@@ -35,6 +36,7 @@ use std::{
         unix::ffi::OsStrExt,
     },
     ptr::NonNull,
+    sync::LazyLock,
 };
 
 use bitflags::bitflags;
@@ -42,7 +44,7 @@ use btoi::{btoi, btoi_radix};
 use indexmap::IndexSet;
 use nix::{
     errno::Errno,
-    fcntl::{open, readlinkat, OFlag},
+    fcntl::{open, readlinkat, OFlag, ResolveFlag},
     libc::pid_t,
     sys::{
         stat::Mode,
@@ -51,7 +53,6 @@ use nix::{
     },
     unistd::{getpid, lseek64, read, write, Gid, Pid, Uid, Whence},
 };
-use once_cell::sync::Lazy;
 use procfs_core::{
     net::UnixNetEntries,
     process::{LimitValue, MMPermissions, MMapPath, MemoryMaps, Namespace, Namespaces},
@@ -362,10 +363,23 @@ pub struct Status {
     pub sig_caught: SydSigSet,
 }
 
-/// Safe `OFlags` to use when opening `/proc` files.
+/// Safe `OFlags` to use when opening proc(5) files.
 pub const OFLAG_PROC: OFlag =
     OFlag::from_bits_retain(libc::O_RDONLY | libc::O_CLOEXEC | libc::O_NOFOLLOW | libc::O_NOCTTY);
 
+/// Safe `ResolveFlag` to use when opening proc(5) files.
+pub const RFLAG_PROC: ResolveFlag = ResolveFlag::from_bits_retain(
+    libc::RESOLVE_NO_MAGICLINKS
+        | libc::RESOLVE_NO_SYMLINKS
+        | libc::RESOLVE_BENEATH
+        | libc::RESOLVE_NO_XDEV,
+);
+
+/// Function to safely open a proc(5) file and return an `OwnedFd`.
+pub fn safe_open_proc(path: &XPath) -> Result<OwnedFd, Errno> {
+    safe_open(PROC_FILE(), path, OFLAG_PROC, RFLAG_PROC)
+}
+
 // major 204 is a mess -- "Low-density serial ports"
 const LOW_DENSITY_NAMES: &[&str] = &[
     "LU0", "LU1", "LU2", "LU3", "FB0", "SA0", "SA1", "SA2", "SC0", "SC1", "SC2", "SC3", "FW0",
@@ -431,7 +445,7 @@ pub fn proc_stat(pid: Pid) -> Result<Stat, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"stat");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -444,7 +458,7 @@ pub fn proc_statm(pid: Pid) -> Result<Statm, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"statm");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -463,7 +477,7 @@ pub fn proc_status_open(pid: Pid) -> Result<OwnedFd, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"status");
 
-    safe_open(PROC_FILE(), &path, OFLAG_PROC).or(Err(Errno::ESRCH))
+    safe_open_proc(&path)
 }
 
 /// Return the status of the given process from the given Fd.
@@ -516,7 +530,7 @@ pub fn proc_stack_pointer(pid: Pid) -> Result<u64, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"syscall");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .map(File::from)
         .map(BufReader::new)
         .or(Err(Errno::ESRCH))?;
@@ -548,7 +562,7 @@ pub fn proc_auxv(pid: Pid) -> Result<SydHashMap<u64, u64>, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"auxv");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -595,7 +609,7 @@ pub fn proc_comm(pid: Pid) -> Result<XPathBuf, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"comm");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .or(Err(Errno::ESRCH))
         .map(File::from)?;
 
@@ -633,7 +647,7 @@ pub fn proc_cmdline(pid: Pid) -> Result<XPathBuf, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"cmdline");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .or(Err(Errno::ESRCH))
         .map(File::from)?;
 
@@ -693,7 +707,7 @@ pub fn proc_maps_open(pid: Pid) -> Result<OwnedFd, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"maps");
 
-    safe_open(PROC_FILE(), &path, OFLAG_PROC).or(Err(Errno::ESRCH))
+    safe_open_proc(&path)
 }
 
 /// Return the memory maps of the given process from the given Fd.
@@ -710,7 +724,7 @@ pub fn proc_smaps(pid: Pid) -> Result<Vec<SydMemoryMap>, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"smaps");
 
-    let reader = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let reader = safe_open_proc(&path)
         .map(File::from)
         .map(BufReader::new)
         .or(Err(Errno::ESRCH))?;
@@ -744,11 +758,7 @@ pub fn proc_pipemax() -> Result<libc::c_int, Errno> {
 
 /// Open a FD to /proc/sys/fs/pipe-max-size and return an OwnedFd.
 pub fn proc_pipemax_open() -> Result<OwnedFd, Errno> {
-    safe_open(
-        PROC_FILE(),
-        XPath::from_bytes(b"sys/fs/pipe-max-size"),
-        OFLAG_PROC,
-    )
+    safe_open_proc(XPath::from_bytes(b"sys/fs/pipe-max-size"))
 }
 
 /// Retrieve the system pipe max limit from the given Fd.
@@ -1328,7 +1338,7 @@ pub fn proc_environ_open(pid: Pid) -> Result<OwnedFd, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"environ");
 
-    safe_open(PROC_FILE(), &path, OFLAG_PROC).or(Err(Errno::ESRCH))
+    safe_open_proc(&path)
 }
 
 /// Parses /proc/pid/environ from given Fd and returns a HashMap.
@@ -1420,7 +1430,7 @@ pub fn proc_crypto() -> Result<CryptoTable, Errno> {
 
 /// Open a FD to _proc_crypto_(5) and return an OwnedFd.
 pub fn proc_crypto_open() -> Result<OwnedFd, Errno> {
-    safe_open(PROC_FILE(), XPath::from_bytes(b"crypto"), OFLAG_PROC)
+    safe_open_proc(XPath::from_bytes(b"crypto"))
 }
 
 /// Parses /proc/crypto from given Fd and returns the table of crypto algorithms.
@@ -1431,9 +1441,7 @@ pub fn proc_crypto_read(fd: OwnedFd) -> Result<CryptoTable, Errno> {
 /// Parses /proc/net/unix and returns the inodes of all UNIX domain sockets.
 pub fn proc_unix_get_inodes() -> Result<SydHashSet<u64>, Errno> {
     let path = XPath::from_bytes(b"net/unix");
-    let file = safe_open(PROC_FILE(), path, OFLAG_PROC)
-        .map(File::from)
-        .map(BufReader::new)?;
+    let file = safe_open_proc(path).map(File::from).map(BufReader::new)?;
     Ok(UnixNetEntries::from_buf_read(file)
         .map(|e| e.0)
         .or(Err(Errno::EPERM))?
@@ -1470,10 +1478,10 @@ pub fn proc_unix_get_inodes() -> Result<SydHashSet<u64>, Errno> {
 pub fn proc_set_at_secure(pid: Pid, elf_type: ElfType) -> Result<(), Errno> {
     // This function requires CONFIG_CROSS_MEMORY_ATTACH,
     // unless Syd is explicitly configured to fallback.
-    static FORCE_PROC: Lazy<bool> =
-        Lazy::new(|| std::env::var_os(ENV_NO_CROSS_MEMORY_ATTACH).is_some());
-    static PROC_MEM_F: Lazy<bool> =
-        Lazy::new(|| std::env::var_os(ENV_PROC_PID_MEM_FALLBACK).is_some());
+    static FORCE_PROC: LazyLock<bool> =
+        LazyLock::new(|| std::env::var_os(ENV_NO_CROSS_MEMORY_ATTACH).is_some());
+    static PROC_MEM_F: LazyLock<bool> =
+        LazyLock::new(|| std::env::var_os(ENV_PROC_PID_MEM_FALLBACK).is_some());
     proc_set_at_secure2(
         pid,
         elf_type,
@@ -1501,7 +1509,7 @@ pub fn proc_set_at_secure2(
         let mut flags = OFLAG_PROC;
         flags.remove(OFlag::O_RDONLY);
         flags.insert(OFlag::O_RDWR);
-        match safe_open(PROC_FILE(), &path, flags) {
+        match safe_open(PROC_FILE(), &path, flags, RFLAG_PROC) {
             Ok(fd) => Some(File::from(fd)),
             Err(_) => return Err(Errno::EACCES),
         }
@@ -1788,7 +1796,12 @@ pub fn proc_pid_fd(pid: Option<Pid>) -> Result<Vec<(RawFd, XPathBuf)>, Errno> {
     let mut dir = XPathBuf::from_pid(pid);
     dir.push(b"fd");
 
-    let dir = safe_open(PROC_FILE(), &dir, OFLAG_PROC | OFlag::O_DIRECTORY)?;
+    let dir = safe_open(
+        PROC_FILE(),
+        &dir,
+        OFLAG_PROC | OFlag::O_DIRECTORY,
+        RFLAG_PROC,
+    )?;
 
     let mut dot = 0u8;
     let mut res = vec![];
@@ -1820,7 +1833,7 @@ pub fn proc_max_open_files(pid: Pid) -> Result<LimitValue, Errno> {
     let mut path = XPathBuf::from_pid(pid);
     path.push(b"limits");
 
-    let mut file = safe_open(PROC_FILE(), &path, OFLAG_PROC)
+    let mut file = safe_open_proc(&path)
         .map(File::from)
         .or(Err(Errno::ESRCH))?;
 
@@ -1880,6 +1893,138 @@ pub fn log_proc_pid_fd(pid: Option<Pid>) -> Result<(), Errno> {
     Ok(())
 }
 
+/// Kernel taint strings.
+///
+/// See: https://docs.kernel.org/admin-guide/tainted-kernels.html#table-for-decoding-tainted-state
+pub static KERNEL_TAINT_STRINGS: [&str; 20] = [
+    "P (Proprietary module was loaded)",
+    "F (Module was force loaded)",
+    "S (Kernel running on an out of specification system)",
+    "R (Module was force unloaded)",
+    "M (Processor reported a Machine Check Exception)",
+    "B (Bad page referenced or unexpected page flags)",
+    "U (Taint requested by userspace application)",
+    "D (Kernel died recently: OOPS/BUG)",
+    "A (ACPI table overridden by user)",
+    "W (Kernel issued warning)",
+    "C (Staging driver was loaded)",
+    "I (Workaround for bug in platform firmware applied)",
+    "O (Externally-built \"out-of-tree\" module was loaded)",
+    "E (Unsigned module was loaded)",
+    "L (Soft lockup occurred)",
+    "K (Kernel has been live patched)",
+    "X (Auxiliary taint, used by distros)",
+    "T (Kernel built with struct randomization plugin)",
+    "N (An in-kernel test has been run)",
+    "J (Userspace used a mutating debug operation in fwctl)",
+];
+
+bitflags! {
+    /// Bitflags for `/proc/sys/kernel/tainted`.
+    #[derive(Clone, Copy, Debug, Eq, PartialEq, Hash)]
+    pub struct KernelTaintFlags: u64 {
+        /// Proprietary module was loaded
+        const PROPRIETARY_MODULE = 1 << 0; // P
+        /// Module was force loaded
+        const MODULE_FORCE_LOADED = 1 << 1; // F
+        /// Kernel running on an out of specification system
+        const OUT_OF_SPEC_SYSTEM = 1 << 2; // S
+        /// Module was force unloaded
+        const MODULE_FORCE_UNLOADED = 1 << 3; // R
+        /// Processor reported a Machine Check Exception
+        const MACHINE_CHECK_EXCEPTION = 1 << 4; // M
+        /// Bad page referenced or unexpected page flags
+        const BAD_PAGE_REFERENCE = 1 << 5; // B
+        /// Taint requested by userspace application
+        const USER_REQUEST = 1 << 6; // U
+        /// Kernel died recently: OOPS/BUG
+        const OOPS_OR_BUG = 1 << 7; // D
+        /// ACPI table overridden by user
+        const ACPI_TABLE_OVERRIDDEN = 1 << 8; // A
+        /// Kernel issued warning
+        const WARNING_ISSUED = 1 << 9; // W
+        /// Staging driver was loaded
+        const STAGING_DRIVER_LOADED = 1 << 10; // C
+        /// Workaround for bug in platform firmware applied
+        const WORKAROUND_FW_BUG = 1 << 11; // I
+        /// Externally-built "out-of-tree" module was loaded
+        const OUT_OF_TREE_MODULE = 1 << 12; // O
+        /// Unsigned module was loaded
+        const UNSIGNED_MODULE_LOADED = 1 << 13; // E
+        /// Soft lockup occurred
+        const SOFT_LOCKUP_OCCURRED = 1 << 14; // L
+        /// Kernel has been live patched
+        const LIVE_PATCHED = 1 << 15; // K
+        /// Auxiliary taint, used by distros
+        const AUXILIARY = 1 << 16; // X
+        /// Kernel built with struct randomization plugin
+        const BUILT_WITH_RANDOMIZATION = 1 << 17; // T
+        /// An in-kernel test has been run
+        const IN_KERNEL_TEST_RUN = 1 << 18; // N
+        /// Userspace used a mutating debug operation in fwctl
+        const FWCTL_DEBUG_WRITE_USED = 1 << 19; // J
+    }
+}
+
+impl KernelTaintFlags {
+    /// Returns taint reasons (verbatim) for set bits 0..=19.
+    pub fn reasons(self) -> impl Iterator<Item = &'static str> + 'static {
+        (0..KERNEL_TAINT_STRINGS.len())
+            .filter(move |bit| self.bits() & (1u64 << bit) != 0)
+            .map(|bit| KERNEL_TAINT_STRINGS[bit])
+    }
+
+    /// True if *no* bits are set (bitfield == 0).
+    pub fn is_clean(self) -> bool {
+        self.is_empty()
+    }
+}
+
+impl fmt::Display for KernelTaintFlags {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        if self.is_clean() {
+            write!(f, "Kernel is not tainted (0).")
+        } else {
+            write!(
+                f,
+                "Kernel is tainted: {}.",
+                self.reasons().collect::<Vec<_>>().join(", ")
+            )
+        }
+    }
+}
+
+/// Retrieves the value of /proc/sys/kernel/tainted as KernelTaintFlags.
+#[allow(clippy::disallowed_methods)]
+pub fn proc_kernel_taint() -> Result<KernelTaintFlags, Errno> {
+    // SAFETY: Informational command,
+    // runs before syd::config::proc_init.
+    #[allow(clippy::disallowed_methods)]
+    let fd = nix::fcntl::openat(
+        AT_BADFD,
+        XPath::from_bytes(b"/proc/sys/kernel/tainted"),
+        OFLAG_PROC,
+        Mode::empty(),
+    )?;
+
+    // Read up to 24 bytes + 1 byte for \n.
+    let mut data = [0u8; 25];
+    let mut nread = 0usize;
+    while nread < data.len() {
+        #[allow(clippy::arithmetic_side_effects)]
+        match read(&fd, &mut data[nread..]) {
+            Ok(0) => break,
+            Ok(n) => nread += n,
+            Err(Errno::EINTR) => continue,
+            Err(_) => return Err(Errno::ESRCH),
+        }
+    }
+
+    // Parse decimal -> u64, then map to flags (retain unknown higher bits).
+    let val = btoi::<u64>(trim_ascii(&data[..nread])).or(Err(Errno::EINVAL))?;
+    Ok(KernelTaintFlags::from_bits_retain(val))
+}
+
 #[cfg(test)]
 mod tests {
     use std::os::unix::fs::PermissionsExt;
