@@ -13,8 +13,8 @@
 //! Main entry point.
 
 // We like clean and simple code with documentation.
+// Keep in sync with lib.rs.
 #![deny(missing_docs)]
-#![deny(clippy::allow_attributes_without_reason)]
 #![deny(clippy::arithmetic_side_effects)]
 #![deny(clippy::as_ptr_cast_mut)]
 #![deny(clippy::as_underscore)]
@@ -52,6 +52,7 @@
 
 use std::{
     env,
+    env::VarError,
     ffi::{CString, OsStr, OsString},
     fs::{File, OpenOptions},
     io::{BufWriter, Write},
@@ -65,6 +66,7 @@ use std::{
     str::FromStr,
 };
 
+use data_encoding::HEXLOWER;
 use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
 use nix::{
     errno::Errno,
@@ -86,21 +88,24 @@ use syd::{
     caps,
     compat::{waitid, WaitStatus},
     config::*,
-    confine::{nsflags_name, ExportMode},
+    confine::ExportMode,
     err::err2no,
     error,
-    fs::{closeexcept, pidfd_open, retry_on_eintr, set_cloexec},
-    hash::{SydHashMap, SydHashSet},
+    fs::{closeexcept, format_clone_flags, format_clone_names, pidfd_open, set_cloexec},
+    hash::{get_at_random_hex, hash, HashAlgorithm, SydHashMap, SydHashSet},
     hook::Supervisor,
-    info,
+    ignore_signals, info,
     landlock_policy::LandlockPolicy,
     log::log_init,
     loopback_set_up,
     path::XPathBuf,
+    printf,
     sandbox::Sandbox,
+    seal::ensure_sealed,
+    syd_code_name, syd_info,
     syslog::LogLevel,
     unshare::{GidMap, UidMap},
-    warn,
+    warn, IgnoreSignalOpts,
 };
 
 // Set global allocator to tcmalloc if profiling is enabled.
@@ -111,8 +116,8 @@ static GLOBAL: tcmalloc::TCMalloc = tcmalloc::TCMalloc;
 syd::main! {
     use lexopt::prelude::*;
 
-    // Set SIGPIPE handler to default.
-    syd::set_sigpipe_dfl()?;
+    // Initialize logging..
+    log_init(LogLevel::Warn, Some(libc::STDERR_FILENO))?;
 
     // Parse CLI options.
     //
@@ -137,16 +142,20 @@ syd::main! {
                         help();
                         return Ok(ExitCode::SUCCESS);
                     }
+                    "-C" | "--check" => {
+                        syd_info(true)?;
+                        return Ok(ExitCode::SUCCESS);
+                    }
                     "-V" | "--version" => {
-                        syd::syd_info();
+                        syd_info(false)?;
                         return Ok(ExitCode::SUCCESS);
                     }
                     "--el" => {
-                        println!("{SYD_EL}");
+                        printf!("{SYD_EL}")?;
                         return Ok(ExitCode::SUCCESS);
                     }
                     "--sh" => {
-                        println!("{ESYD_SH}");
+                        printf!("{ESYD_SH}")?;
                         return Ok(ExitCode::SUCCESS);
                     }
                     "-q" => is_quick = true,
@@ -162,39 +171,112 @@ syd::main! {
     // - https://github.com/opencontainers/runc/commit/0a8e4117e7f715d5fbeef398405813ce8e88558b
     // - https://github.com/lxc/lxc/commit/6400238d08cdf1ca20d49bafb85f4e224348bf9d
     // Note: syd's procfs protections is another layer of defense against this.
-    if !is_quick {
-        syd::seal::ensure_sealed()?;
+    #[allow(clippy::disallowed_methods)]
+    let cookie = if !is_quick {
+        match env::var(ENV_RAND) {
+            Ok(cookie0) => {
+                // Best-effort ensure cookie0 was not tampered.
+                assert_eq!(cookie0.len(), 32,
+                    "PANIC: Internal environment variable {ENV_RAND} tampered by user!");
+                assert!(cookie0.bytes().all(|b| b.is_ascii_hexdigit() && !b.is_ascii_uppercase()),
+                    "PANIC: Internal environment variable {ENV_RAND} tampered by user!");
+                let cookie1 = get_at_random_hex(false);
+                env::set_var(ENV_RAND, format!("{cookie0}{cookie1}"));
+                info!("ctx": "set_random_cookie",
+                    "cookie": [&cookie0, &cookie1], "src": "AT_RANDOM",
+                    "msg": format!("appended random cookie from AT_RANDOM {cookie0}+{cookie1}={cookie0}{cookie1} after memfd-reexec"));
+            }
+            Err(VarError::NotPresent) => {
+                let cookie = get_at_random_hex(false);
+                env::set_var(ENV_RAND, &cookie);
+                info!("ctx": "set_random_cookie",
+                    "cookie": &cookie, "src": "AT_RANDOM",
+                    "msg": format!("set random cookie from AT_RANDOM to {cookie}"));
+            }
+            Err(VarError::NotUnicode(cookie)) => {
+                error!("ctx": "set_random_cookie",
+                    "cookie": &cookie, "src": "AT_RANDOM", "err": libc::EINVAL,
+                    "msg": format!("get random cookie from {ENV_RAND} failed: {}", Errno::EINVAL));
+            }
+        }
+
+        match ensure_sealed() {
+            Ok(()) => env::var(ENV_RAND).unwrap(),
+            Err(errno) => {
+                error!("ctx": "memfd_reexec",
+                    "err": errno as i32,
+                    "msg": format!("reexecute self with a sealed memfd failed: {errno}"),
+                    "tip": "set SYD_QUICK_BOOT and/or submit a bug report");
+                return Err(errno.into());
+            }
+        }
+    } else {
+        env::set_var(ENV_RAND, get_at_random_hex(false));
+        env::var(ENV_RAND).unwrap()
+    };
+
+    // Generate unique sandbox id from AT_RANDOM bytes.
+    // Allow the user to override by setting SYD_ID.
+    // Panic if SYD_ID is incorrectly formatted.
+    #[allow(clippy::disallowed_methods)]
+    if let Some(sandbox_id) = env::var_os(ENV_ID) {
+        assert_eq!(sandbox_id.len(), 128,
+            "PANIC: Sandbox ID in SYD_ID environment variable isn't in correct format!");
+        assert!(sandbox_id.as_bytes().iter().all(|b| b.is_ascii_hexdigit() && !b.is_ascii_uppercase()),
+            "PANIC: Sandbox ID in SYD_ID environment variable isn't in correct format!");
+        let machine_id = &sandbox_id.as_bytes()[..32];
+        assert!(machine_id.iter().any(|&b| b != b'0'),
+            "PANIC: Sandbox ID in SYD_ID environment variable isn't in correct format!");
+    } else {
+        let sandbox_id = HEXLOWER.encode(&hash(cookie.as_bytes(), HashAlgorithm::Sha512).unwrap());
+        env::set_var(ENV_ID, &sandbox_id);
+        info!("ctx": "set_sandbox_id",
+            "id": &sandbox_id, "cookie": &cookie, "hash": "sha3-512",
+            "msg": format!("generated syd id:{sandbox_id} from cookie:{cookie} using SHA3-512"));
     }
 
     // SYD_PID_FN -> Write PID file.
-    if let Some(path) = env::var_os(ENV_PID_FN) {
-        let mut pid = itoa::Buffer::new();
-        let pid_str = pid.format(getpid().as_raw());
+    if let Some(pid_fn) = env::var_os(ENV_PID_FN).map(XPathBuf::from) {
+        let pid = getpid().as_raw();
 
-        #[allow(clippy::disallowed_methods)]
-        let pid_file = OpenOptions::new()
-            .create_new(true)
-            .write(true)
+        let mut pid_str = itoa::Buffer::new();
+        let pid_str = pid_str.format(pid);
+
+        let mut openopts = OpenOptions::new();
+        openopts
             .mode(0o400)
-            .open(path)?;
-        let mut pid_file = BufWriter::new(pid_file);
+            .write(true)
+            .create_new(true);
+        #[allow(clippy::disallowed_methods)]
+        let mut pid_file = match openopts.open(&pid_fn).map(BufWriter::new) {
+            Ok(pid_file) => pid_file,
+            Err(error) => {
+                let errno = err2no(&error);
+                error!("ctx": "write_pid_file",
+                    "pid_file": &pid_fn, "err": errno as i32,
+                    "msg": format!("pid file create error: {error}"),
+                    "tip": format!("remove file `{pid_fn}' or unset SYD_PID_FN"));
+                return Err(error.into());
+            }
+        };
 
-        pid_file.write_all(pid_str.as_bytes())?;
+        match pid_file.write_all(pid_str.as_bytes()) {
+            Ok(_) => {
+                info!("ctx": "write_pid_file",
+                    "msg": format!("Syd pid {pid} written to file `{pid_fn}'"),
+                    "pid_file": &pid_fn);
+            }
+            Err(error) => {
+                let errno = err2no(&error);
+                error!("ctx": "write_pid_file",
+                    "pid_fn": &pid_fn, "err": errno as i32,
+                    "msg": format!("pid file write error: {error}"),
+                    "tip": format!("remove file `{pid_fn}' or unset SYD_PID_FN"));
+                return Err(error.into());
+            }
+        }
     }
 
-    // Ignore all signals except SIG{KILL,STOP,PIPE,CHLD,Core}.
-    syd::ignore_signals()?;
-
-    // Initialize logging..
-    log_init(
-        if is_login {
-            LogLevel::Info
-        } else {
-            LogLevel::Warn
-        },
-        Some(libc::STDERR_FILENO),
-    )?;
-
     // Parse CLI arguments
     let mut export: Option<ExportMode> = ExportMode::from_env();
     let mut sandbox: Sandbox = Sandbox::default();
@@ -254,10 +336,17 @@ syd::main! {
                 help();
                 return Ok(ExitCode::SUCCESS);
             }
-            Short('V') | Long("version") => {
-                syd::syd_info();
+            Short('C') | Long("check") => {
+                syd_info(true)?;
                 return Ok(ExitCode::SUCCESS);
             }
+            // syd -V is called often by paludis.
+            // We want to keep its output short and parseable.
+            Short('V') | Long("version") => {
+                syd_info(false)?;
+                return Ok(ExitCode::SUCCESS);
+            }
+            Short('v') | Long("verbose") => sandbox.verbose = true,
             Long("el") => {
                 println!("{SYD_EL}");
                 return Ok(ExitCode::SUCCESS);
@@ -426,6 +515,28 @@ syd::main! {
         }
     };
 
+    // Ignore all signals except the following signals:
+    // SIGALRM, SIGCHLD, SIGKILL, SIGSTOP.
+    // Skip ignoring signals with default action Core,
+    // if trace/allow_unsafe_prlimit:1 is set at startup.
+    let mut opts = IgnoreSignalOpts::SkipIgnoreAlarm;
+    if sandbox.flags.allow_unsafe_prlimit() {
+        opts.insert(IgnoreSignalOpts::SkipIgnoreCoreDump);
+    }
+    match ignore_signals(opts) {
+        Ok(()) => {
+            info!("ctx": "ignore_signals",
+                "opt": opts, "msg": "ignored signals");
+        }
+        Err(errno) => {
+            error!("ctx": "ignore_signals",
+                "opt": opts, "err": errno as i32,
+                "msg": format!("ignoring signals failed: {errno}"),
+                "tip": "check with SYD_LOG=debug and/or submit a bug report");
+            return Err(errno.into());
+        }
+    }
+
     // SAFETY: We cannot support NEWPID without NEWNS.
     // ie, pid namespace must have its own private /proc.
     if sandbox.flags.unshare_pid() {
@@ -471,8 +582,7 @@ syd::main! {
             })
             .collect();
 
-        // Step 5: Spawn syd-pty process with a double-fork.
-        // Pass PTY main end to it.
+        // Step 5: Spawn syd-pty process, and pass PTY main end to it.
         let mut cmd = std::process::Command::new(
             env::var_os("CARGO_BIN_EXE_syd-pty").unwrap_or(OsString::from("syd-pty")),
         );
@@ -494,26 +604,13 @@ syd::main! {
         if !pty_debug {
             cmd.stderr(Stdio::null());
         }
-        // SAFETY: In libc we trust.
-        match unsafe { fork() }? {
-            ForkResult::Parent { child } => {
-                retry_on_eintr(|| {
-                    waitid(
-                        Id::Pid(child),
-                        WaitPidFlag::WEXITED | WaitPidFlag::__WNOTHREAD,
-                    )
-                })?;
-            }
-            ForkResult::Child => {
-                // SAFETY: In libc we trust.
-                match unsafe { fork() }? {
-                    ForkResult::Parent { .. } => {
-                        // SAFETY: In libc we trust.
-                        unsafe { libc::_exit(0) };
-                    }
-                    ForkResult::Child => return Err(cmd.exec().into()),
-                };
-            }
+        if let Err(error) = cmd.spawn() {
+            let errno = err2no(&error);
+            error!("ctx": "spawn_pty",
+                "msg": format!("syd-pty spawn error: {error}"),
+                "tip": "put syd-pty(1) in PATH or set CARGO_BIN_EXE_syd-pty",
+                "err": errno as i32);
+            return Err(error.into());
         }
         drop(pidfd);
         drop(pty_m);
@@ -594,7 +691,14 @@ syd::main! {
         if !proxy_debug {
             cmd.stderr(Stdio::null());
         }
-        cmd.spawn()?;
+        if let Err(error) = cmd.spawn() {
+            let errno = err2no(&error);
+            error!("ctx": "spawn_tor",
+                "msg": format!("syd-tor spawn error: {error}"),
+                "tip": "put syd-tor(1) in PATH or set CARGO_BIN_EXE_syd-tor",
+                "err": errno as i32);
+            return Err(error.into());
+        }
         drop(pidfd);
 
         let proxy_repr = sandbox
@@ -640,12 +744,6 @@ syd::main! {
                 Ok(ExitCode::from(u8::try_from(errno as i32).unwrap_or(127)))
             }
         };
-    } else {
-        let n = nsflags_name(namespaces);
-        let s = if n.len() > 1 { "s" } else { "" };
-        let n = n.join(", ");
-        info!("ctx": "setup_container",
-            "msg": format!("setting up Syd container with {n} namespace{s}"));
     }
 
     let id_buf = if sandbox.flags.unshare_user() {
@@ -704,7 +802,23 @@ syd::main! {
 
     // CLONE_NEWTIME may only be used with unshare(2).
     // CloneFlags don't support CLONE_NEWTIME directly so we use retain.
-    unshare(CloneFlags::from_bits_retain(namespaces))?;
+    let clone_flags = CloneFlags::from_bits_retain(namespaces);
+    let clone_names = format_clone_flags(clone_flags);
+    let clone_types = format_clone_names(&clone_names);
+    match unshare(clone_flags) {
+        Ok(()) => {
+            info!("ctx": "unshare_namespaces",
+                "ns": clone_names,
+                "msg": format!("unshared into {clone_types}"));
+        }
+        Err(errno) => {
+            error!("ctx": "unshare_namespaces",
+                "ns": clone_names, "err": errno as i32,
+                "msg": format!("unshare into {clone_types} failed: {errno}"),
+                "tip": "check with SYD_LOG=debug and/or set `unshare/user:1'");
+            return Err(errno.into());
+        }
+    }
 
     // Write uid/gid map for user namespace.
     #[allow(clippy::disallowed_methods)]
@@ -1002,5 +1116,49 @@ syd::main! {
 }
 
 fn help() {
-    println!("syd [-acefhlmpqxEPV] [--] {{command [arg...]}}");
+    let mut c_blue = "\x1b[0;1;35;95m";
+    let mut c_bold = "\x1b[1m";
+    let mut c_cyan = "\x1b[0;1;36;96m";
+    let mut c_green = "\x1b[0;1;32;92m";
+    let mut c_orng = "\x1b[0;1;34;94m";
+    let mut c_red = "\x1b[0;1;31;91m";
+    let mut c_res = "\x1b[0m";
+    let mut c_yll = "\x1b[0;1;33;93m";
+    if !isatty(std::io::stdout()).unwrap_or(false) {
+        c_blue = "";
+        c_bold = "";
+        c_cyan = "";
+        c_green = "";
+        c_orng = "";
+        c_red = "";
+        c_res = "";
+        c_yll = "";
+    }
+
+    println!(
+        "{c_red}syd{c_res} {c_cyan}{}{c_res} ({c_orng}{}{c_res})",
+        *syd::config::VERSION,
+        syd_code_name()
+    );
+    println!("{c_yll}Rock solid application kernel{c_res}");
+    println!("{c_blue}Author:{c_res} {c_yll}Ali Polatel{c_res} <{c_bold}alip@chesswob.org{c_res}>");
+    println!("{c_blue}License:{c_res} {c_yll}GPL-3.0-only{c_res}");
+    println!();
+    println!("{c_green}$ syd [-acefhlmpqxEPV] [--] {{command [arg...]}}{c_res}");
+    println!("  {c_bold}Run a program under syd.{c_res}");
+    println!("{c_green}$ syd [-acefhlmpqxEPV] [--] {{library.so}}{c_res}");
+    println!("  {c_bold}Load a library under syd.{c_res}");
+    println!("{c_green}$ syd --api{c_res}");
+    println!("  {c_bold}Print syd(2) API specification.{c_res}");
+    println!("{c_green}$ syd --check{c_res}");
+    println!("  {c_bold}Print sandboxing support information.{c_res}");
+    println!("{c_green}$ syd --el{c_res}");
+    println!("  {c_bold}Output syd.el the Emacs Lisp implementation of syd(2) interface.{c_res}");
+    println!("{c_green}$ syd --sh{c_res}");
+    println!("  {c_bold}Output a shell script which defines the esyd helper function.{c_res}");
+    println!();
+    print!("{SEE_EMILY_PLAY}");
+    println!();
+    println!("{c_orng}Send bug reports to{c_res} {c_bold}https://gitlab.exherbo.org/groups/sydbox/-/issues{c_res}");
+    println!("{c_orng}Attaching poems encourages consideration tremendously.{c_res}");
 }
