@@ -151,6 +151,7 @@ fn list() {
     println!("landlock");
     println!("lib");
     println!("linux");
+    println!("ltp");
     println!("noipv4");
     println!("noipv6");
     println!("nomem");
@@ -208,6 +209,15 @@ fn dump(name: &str) -> SydResult<()> {
                 println!("{command}");
             }
         }
+        "ltp" => {
+            println!("# Syd profile: LTP");
+            println!("# Number of rules: {}", syd::config::PROFILE_LTP.len());
+            println!("# Copyright (c) 2025 Ali Polatel <alip@chesswob.org>");
+            println!("# SPDX-License-Identifier: GPL-3.0");
+            for command in syd::config::PROFILE_LTP {
+                println!("{command}");
+            }
+        }
         "kvm" => {
             println!("# Syd profile: KVM");
             println!("# Number of rules: {}", syd::config::PROFILE_KVM.len());
