@@ -45,6 +45,7 @@ syd::main! {
             println!("Available sets are:");
             println!("- cpu");
             println!("- deny");
+            println!("- deprecated");
             println!("- ebpf");
             println!("- futex");
             println!("- hook");
@@ -92,6 +93,11 @@ syd::main! {
                 println!("{name}");
             }
         }
+        Some("deprecated") => {
+            for name in syd::config::DEPRECATED_SYSCALLS {
+                println!("{name}");
+            }
+        }
         Some("ebpf") => {
             for name in syd::config::EBPF_SYSCALLS {
                 println!("{name}");
