@@ -80,10 +80,12 @@ use syd::{
     compat::{waitid, WaitStatus},
     err::{SydError, SydResult},
     hook::Supervisor,
+    ignore_signals,
     log::log_init,
     path::XPath,
     sandbox::Sandbox,
     syslog::LogLevel,
+    IgnoreSignalOpts,
 };
 use tabwriter::TabWriter;
 use tracing_subscriber::layer::SubscriberExt;
@@ -177,9 +179,13 @@ impl Executor for SydExecutor {
             serde_json::to_string(&sandbox).unwrap_or("?".to_string())
         );
 
-        // Ignore all signals except SIG{KILL,STOP,PIPE,CHLD,Core}.
+        // Ignore all signals except SIG{CHLD,KILL,STOP}.
         // This is used to ensure we can deny {rt_,}sigreturn(2) to mitigate SROP.
-        syd::ignore_signals().or(Err(ExecutorError::CantHandle("Failed to ignore signals!")))?;
+        let mut opts = IgnoreSignalOpts::SkipIgnoreAlarm;
+        if sandbox.flags.allow_unsafe_prlimit() {
+            opts.insert(IgnoreSignalOpts::SkipIgnoreCoreDump);
+        }
+        ignore_signals(opts).or(Err(ExecutorError::CantHandle("Failed to ignore signals!")))?;
 
         // Run command under the Syd sandbox and exit with the return code.
         #[allow(clippy::disallowed_methods)]
