@@ -32,11 +32,13 @@ use syd::{
     confine::{confine_mdwe, extend_ioctl},
     err::SydResult,
     fs::{closeexcept, duprand, set_exclusive, set_nonblock, winsize_get, winsize_set},
+    ignore_signals,
     ioctl::{
         TCGETS, TCGETS2, TCSETS, TCSETS2, TCSETSF, TCSETSF2, TCSETSW, TCSETSW2, TIOCGWINSZ,
         TIOCSWINSZ,
     },
     landlock_policy::LandlockPolicy,
+    IgnoreSignalOpts,
 };
 
 // This is from <linux/tty.h>
@@ -50,9 +52,9 @@ syd::main! {
     // Parse options.
     let opts = parse_options()?;
 
-    // Ignore all signals except SIG{KILL,STOP,PIPE,CHLD,Core}.
+    // Ignore all signals except SIG{CHLD,KILL,STOP}.
     // This is used to ensure we can deny {rt_,}sigreturn(2) to mitigate SROP.
-    syd::ignore_signals()?;
+    ignore_signals(IgnoreSignalOpts::empty())?;
 
     // Close all file descriptors, except:
     // 1. Standard input, output, and error.
@@ -479,7 +481,7 @@ fn confine(sig_fd: RawFd, dry_run: bool, print_rules: bool) -> SydResult<()> {
     chdir("/proc/self/fdinfo")?;
 
     if !dry_run {
-        // SAFETY: Default panic hook wont play well with seccomp.
+        // SAFETY: Default panic hook won't play well with seccomp.
         std::panic::set_hook(Box::new(|_| {}));
 
         // Set up namespace isolation for all available namespaces.
