@@ -10,7 +10,6 @@
 // SPDX-License-Identifier: GPL-3.0
 
 use std::{
-    collections::{HashMap, HashSet},
     env,
     io::stderr,
     net::{IpAddr, Ipv4Addr, SocketAddr},
@@ -32,10 +31,11 @@ use nix::{
     sys::{
         epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
         mman::{mmap_anonymous, MapFlags, ProtFlags},
+        prctl::set_dumpable,
         resource::{getrlimit, setrlimit, Resource},
         socket::{
-            accept4, connect, listen, shutdown, socket, AddressFamily, Backlog, Shutdown, SockFlag,
-            SockType, SockaddrLike, SockaddrStorage, UnixAddr,
+            accept4, listen, shutdown, AddressFamily, Backlog, Shutdown, SockFlag, SockType,
+            SockaddrLike, SockaddrStorage, UnixAddr,
         },
         stat::Mode,
     },
@@ -45,15 +45,22 @@ use sendfd::RecvWithFd;
 use syd::{
     compat::epoll_ctl_safe,
     config::PIPE_BUF,
+    confine::{confine_mdwe, seccomp_native_has_socketcall},
+    cookie::{
+        safe_connect, safe_socket, CONNECT_COOKIE_ARG3, CONNECT_COOKIE_ARG4, CONNECT_COOKIE_ARG5,
+        SOCKET_COOKIE_ARG3, SOCKET_COOKIE_ARG4, SOCKET_COOKIE_ARG5,
+    },
     err::{SydError, SydResult},
     fs::{closeexcept, duprand, fillrandom, randint, retry_on_eintr},
-    hash::SydRandomState,
+    hash::SydHashMap,
+    ignore_signals,
     landlock_policy::LandlockPolicy,
+    parsers::sandbox::PortSet,
     path::XPathBuf,
     sealbox::{
         check_madvise_guard_support, madvise_guard_install, mprotect_none, mprotect_readonly, mseal,
     },
-    seccomp_native_has_socketcall,
+    IgnoreSignalOpts,
 };
 
 // Pong reply to Syd for debugging.
@@ -64,9 +71,9 @@ syd::main! {
     // Parse options.
     let mut opts = parse_options()?;
 
-    // Ignore all signals except SIG{KILL,STOP,PIPE,CHLD,Core}.
+    // Ignore all signals except SIG{CHLD,KILL,STOP}.
     // This is used to ensure we can deny {rt_,}sigreturn(2) to mitigate SROP.
-    syd::ignore_signals()?;
+    ignore_signals(IgnoreSignalOpts::empty())?;
 
     // SAFETY: We leak the UNIX socket fd on purpose,
     // and leave it open for process lifetime.
@@ -229,8 +236,8 @@ syd::main! {
     })
 }
 
-type FdMap = HashMap<RawFd, RawFd, SydRandomState>;
-type CliMap = HashMap<RawFd, Client, SydRandomState>;
+type FdMap = SydHashMap<RawFd, RawFd>;
+type CliMap = SydHashMap<RawFd, Client>;
 
 struct Proxy<'a> {
     debug: bool,
@@ -400,16 +407,18 @@ impl<'a> Proxy<'a> {
         }
 
         // Socket filtering only works if there's no multiplexing socketcall.
+        #[allow(clippy::useless_conversion)]
         if seccomp_native_has_socketcall() {
             for sysname in ["socket", "connect"] {
                 #[allow(clippy::disallowed_methods)]
                 filter.add_rule(ScmpAction::Allow, ScmpSyscall::from_name(sysname).unwrap())?;
             }
         } else {
-            // Restrict socket to the given domain, type and protocol.
+            // Restrict socket(2) to the given domain, type and protocol.
+            // Restrict socket(2) using syscall argument cookies.
             let sock_domain = domain as u64;
             let sock_type = (libc::SOCK_STREAM | libc::SOCK_NONBLOCK | libc::SOCK_CLOEXEC) as u64;
-            let sock_protocol = libc::IPPROTO_IP as u64;
+            let sock_protocol = 0u64;
             #[allow(clippy::disallowed_methods)]
             filter.add_rule_conditional(
                 ScmpAction::Allow,
@@ -418,15 +427,25 @@ impl<'a> Proxy<'a> {
                     scmp_cmp!($arg0 == sock_domain),
                     scmp_cmp!($arg1 == sock_type),
                     scmp_cmp!($arg2 == sock_protocol),
+                    scmp_cmp!($arg3 == (*SOCKET_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*SOCKET_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*SOCKET_COOKIE_ARG5).into()),
                 ],
             )?;
 
-            // Restrict connect to a single safe pointer.
+            // Restrict connect(2) to a single safe pointer.
+            // Restrict connect(2) using syscall argument cookies.
             #[allow(clippy::disallowed_methods)]
             filter.add_rule_conditional(
                 ScmpAction::Allow,
                 ScmpSyscall::from_name("connect").unwrap(),
-                &[scmp_cmp!($arg1 == addr_ptr), scmp_cmp!($arg2 == addr_len)],
+                &[
+                    scmp_cmp!($arg1 == addr_ptr),
+                    scmp_cmp!($arg2 == addr_len),
+                    scmp_cmp!($arg3 == (*CONNECT_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*CONNECT_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*CONNECT_COOKIE_ARG5).into()),
+                ],
             )?;
         }
 
@@ -455,7 +474,7 @@ impl<'a> Proxy<'a> {
         chdir(safe_cwd)?;
 
         if !self.debug {
-            // SAFETY: Default panic hook wont play well with seccomp.
+            // SAFETY: Default panic hook won't play well with seccomp.
             std::panic::set_hook(Box::new(|_| {}));
 
             // Set up namespace isolation.
@@ -487,8 +506,8 @@ impl<'a> Proxy<'a> {
             // Ignore errors as Landlock may not be supported.
             let abi = syd::landlock::ABI::new_current();
             let conn_portset = port.map(|port| {
-                let mut set = HashSet::default();
-                set.insert(port..=port);
+                let mut set = PortSet::with_capacity(0x10000);
+                set.insert(port.into());
                 set
             });
             let policy = LandlockPolicy {
@@ -503,14 +522,10 @@ impl<'a> Proxy<'a> {
 
             // Set up Memory-Deny-Write-Execute protections.
             // Ignore errors as PR_SET_MDWE may not be supported.
-            const PR_SET_MDWE: libc::c_int = 65;
-            const PR_MDWE_REFUSE_EXEC_GAIN: libc::c_ulong = 1;
-            // SAFETY: nix has no wrapper for PR_SET_MDWE.
-            let _ = unsafe { libc::prctl(PR_SET_MDWE, PR_MDWE_REFUSE_EXEC_GAIN, 0, 0, 0) };
+            let _ = confine_mdwe(false);
 
             // Set the process dumpable attribute to not-dumpable.
-            // SAFETY: Our nix version does not have a wrapper for set_dumpable yet.
-            let _ = unsafe { libc::prctl(libc::PR_SET_DUMPABLE, 0, 0, 0, 0) };
+            let _ = set_dumpable(false);
 
             // Deny reading the timestamp counter (x86 only).
             // SAFETY: Our nix version does not have a wrapper for SET_TSC yet.
@@ -725,17 +740,16 @@ impl Client {
         let family = extaddr
             .family()
             .ok_or::<SydError>(Errno::EAFNOSUPPORT.into())?;
-        let fd_ext = socket(
-            family,
-            SockType::Stream,
-            SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC,
-            None,
+        let fd_ext = safe_socket(
+            family as i32,
+            (SockType::Stream as i32) | (SockFlag::SOCK_NONBLOCK | SockFlag::SOCK_CLOEXEC).bits(),
+            0,
         )?;
 
         // Set socket options on the external fd.
         set_socket_options(&fd_ext, None);
 
-        retry_on_eintr(|| connect(fd_ext.as_raw_fd(), extaddr)).or_else(|e| {
+        retry_on_eintr(|| safe_connect(&fd_ext, extaddr)).or_else(|e| {
             if matches!(e, Errno::EAGAIN | Errno::EINPROGRESS) {
                 Ok(())
             } else {
