@@ -40,10 +40,12 @@ use crate::{
     alert,
     config::*,
     confine::{
-        extend_ioctl, scmp_add_setid_rules, scmp_arch_raw, ScmpNotifReq, SydArch, Sydcall,
-        X32_SYSCALL_BIT,
+        confine_scmp_clone, confine_scmp_clone3, extend_ioctl, scmp_add_setid_rules, scmp_arch_raw,
+        ScmpNotifReq, SydArch, Sydcall, EIDRM, EOWNERDEAD, X32_SYSCALL_BIT,
     },
     cookie::{
+        ACCEPT4_COOKIE_ARG4, ACCEPT4_COOKIE_ARG5, BIND_COOKIE_ARG3, BIND_COOKIE_ARG4,
+        BIND_COOKIE_ARG5, CONNECT_COOKIE_ARG3, CONNECT_COOKIE_ARG4, CONNECT_COOKIE_ARG5,
         FTRUNCATE64_COOKIE_ARG4, FTRUNCATE64_COOKIE_ARG5, FTRUNCATE_COOKIE_ARG2,
         FTRUNCATE_COOKIE_ARG3, FTRUNCATE_COOKIE_ARG4, FTRUNCATE_COOKIE_ARG5, LINKAT_COOKIE_ARG5,
         MEMFD_CREATE_COOKIE_ARG2, MEMFD_CREATE_COOKIE_ARG3, MEMFD_CREATE_COOKIE_ARG4,
@@ -51,10 +53,10 @@ use crate::{
         SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG3, SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG4,
         SECCOMP_IOCTL_NOTIF_ADDFD_COOKIE_ARG5, SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG3,
         SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG4, SECCOMP_IOCTL_NOTIF_SEND_COOKIE_ARG5,
-        SOCKET_COOKIE_ARG3, SOCKET_COOKIE_ARG4, SOCKET_COOKIE_ARG5, SYS_SOCKET,
-        TRUNCATE64_COOKIE_ARG4, TRUNCATE64_COOKIE_ARG5, TRUNCATE_COOKIE_ARG2, TRUNCATE_COOKIE_ARG3,
-        TRUNCATE_COOKIE_ARG4, TRUNCATE_COOKIE_ARG5, UNLINKAT_COOKIE_ARG3, UNLINKAT_COOKIE_ARG4,
-        UNLINKAT_COOKIE_ARG5,
+        SOCKET_COOKIE_ARG3, SOCKET_COOKIE_ARG4, SOCKET_COOKIE_ARG5, SYS_ACCEPT4, SYS_BIND,
+        SYS_CONNECT, SYS_SOCKET, TRUNCATE64_COOKIE_ARG4, TRUNCATE64_COOKIE_ARG5,
+        TRUNCATE_COOKIE_ARG2, TRUNCATE_COOKIE_ARG3, TRUNCATE_COOKIE_ARG4, TRUNCATE_COOKIE_ARG5,
+        UNLINKAT_COOKIE_ARG3, UNLINKAT_COOKIE_ARG4, UNLINKAT_COOKIE_ARG5,
     },
     err::{err2no, SydJoinHandle, SydResult},
     fs::{
@@ -67,11 +69,9 @@ use crate::{
     proc::proc_mmap,
     sandbox::{Flags, Sandbox, SandboxGuard},
     timer::AlarmTimer,
-    workers::{aes::AesMap, BindMap, WorkerCache, WorkerData},
+    workers::{aes::AesMap, UnixMap, WorkerCache, WorkerData},
 };
 
-const EOWNERDEAD: i32 = -libc::EOWNERDEAD;
-
 #[derive(Clone)]
 pub(crate) struct Worker {
     fd: RawFd,
@@ -82,8 +82,8 @@ pub(crate) struct Worker {
     keep_alive: Option<u16>,
     should_exit: Arc<AtomicBool>,
     worker_data: Arc<WorkerData>,
-    bind_map: BindMap,
     crypt_map: Option<AesMap>,
+    unix_map: UnixMap,
 }
 
 impl Worker {
@@ -96,8 +96,8 @@ impl Worker {
         keep_alive: Option<u16>,
         should_exit: Arc<AtomicBool>,
         worker_data: Arc<WorkerData>,
-        bind_map: BindMap,
         crypt_map: Option<AesMap>,
+        unix_map: UnixMap,
     ) -> Self {
         let my_sandbox = SandboxGuard::Read(sandbox.read().unwrap_or_else(|err| err.into_inner()));
         let flags = *my_sandbox.flags;
@@ -112,8 +112,8 @@ impl Worker {
             keep_alive,
             should_exit,
             worker_data,
-            bind_map,
             crypt_map,
+            unix_map,
         }
     }
 
@@ -159,7 +159,7 @@ impl Worker {
                 // SAFETY: We use exit_group(2) here to bail,
                 // because expiring idle threads is a critical safety feature.
                 let mut timer = if let Some(keep_alive) = self.keep_alive {
-                    match AlarmTimer::new(keep_alive) {
+                    match AlarmTimer::from_seconds(keep_alive.into()) {
                         Ok(timer) => Some(timer),
                         Err(errno) => {
                             alert!("ctx": "boot", "op": "timer_create_for_emulator_thread",
@@ -229,6 +229,12 @@ impl Worker {
                         sentinel.seccomp_id = None;
                         self.worker_data.decrement_worker_busy();
                     } // else process died-midway, continue.
+
+                    // Exit if there's nothing else to handle.
+                    if self.should_exit.load(Ordering::Relaxed) {
+                        // Time to exit.
+                        break;
+                    }
                 }
 
                 Ok(())
@@ -309,25 +315,34 @@ impl Worker {
             self.fd,
             Arc::clone(&self.cache),
             Arc::clone(&self.sandbox),
-            Arc::clone(&self.bind_map),
             self.crypt_map.as_ref().map(Arc::clone),
+            Arc::clone(&self.unix_map),
         );
-        let response = handler(request);
+        let mut response = handler(request);
 
-        if response.id == 0 && response.val == 0 && response.error == 0 && response.flags == 0 {
-            // All-zero dummy empty seccomp response.
-            // A previous ADDFD request has already replied to the request.
-            // Nothing left to do here.
-            return;
-        } else if response.error == EOWNERDEAD {
-            // EOWNERDEAD is a pseudo errno used by
-            // the stat handler thread to close the
-            // seccomp notify fd upon receiving the
-            // "ghost" command.
-            crate::warn!("ctx": "confine", "op": "enter_ghost_mode", "pid": req.pid,
-                "sys": syscall, "arch": SydArch(req.data.arch), "args": req.data.args,
-                "src": proc_mmap(req.pid(), req.data.instr_pointer).ok());
-        }
+        // Check for the following pseudo errnos:
+        // 1. EIDRM:
+        //    a. A previous SECCOMP_IOCTL_NOTIF_ADDFD has replied to the request with
+        //       SECCOMP_ADDFD_FLAG_SEND already, no need to return a reply again here.
+        //    b. A read-write encryption request was made and the encrypted fd has
+        //       already been returned as a reply with SECCOMP_IOCTL_NOTIF_ADDFD with
+        //       the flag SECCOMP_ADDFD_FLAG_SEND.
+        // 2. EOWNERDEAD: Enter ghost mode.
+        let ghost = match response.error {
+            EIDRM if response.id == 0 && response.val == 0 && response.flags == 0 => return,
+            EOWNERDEAD if response.id == 0 && response.val == 0 && response.flags == 0 => {
+                crate::warn!("ctx": "confine", "op": "enter_ghost_mode", "pid": req.pid,
+                    "sys": syscall, "arch": SydArch(req.data.arch), "args": req.data.args,
+                    "src": proc_mmap(req.pid(), req.data.instr_pointer).ok());
+
+                // Correct the pseudo errno back to success.
+                response.error = 0;
+                response.val = 0;
+
+                true
+            }
+            _ => false,
+        };
 
         let response = libc::seccomp_notif_resp {
             id: response.id,
@@ -336,16 +351,16 @@ impl Worker {
             flags: response.flags,
         };
 
-        // EAGAIN|EINTR is retried.
+        // EINTR is not retried because it may mean child is signaled.
         // ENOENT means child died mid-way.
         // Nothing else we can do on errors here.
         let _ = seccomp_notify_respond(self.fd, std::ptr::addr_of!(response));
 
-        // See above.
-        if response.error == EOWNERDEAD {
+        // Finalize ghost mode as necessary.
+        if ghost {
             let _ = close(self.fd);
 
-            // Inform the monitor thread to exit.
+            // Inform the monitor thread and other emulator threads to exit.
             self.should_exit.store(true, Ordering::Relaxed);
         }
     }
@@ -385,6 +400,11 @@ impl Worker {
         // which we do not need at all.
         // seccomp_add_architectures(&mut ctx)?;
 
+        // Allow clones without namespace flags.
+        confine_scmp_clone(&mut ctx)?;
+        // Deny clone3 with ENOSYS for compatibility.
+        confine_scmp_clone3(&mut ctx)?;
+
         // Deny open and {l,}stat with ENOSYS rather than KillProcess.
         // We need this because std::thread::spawn has unwanted
         // side-effects such as opening /sys/devices/system/cpu/online
@@ -801,9 +821,9 @@ impl Worker {
         }
 
         // socket(2) may be used only with syscall argument cookies.
-        // Note: We only enforce this on architectures where the
-        // system call is direct and there's no socketcall(2)
-        // multiplexer indirection.
+        //
+        // We only enforce this on architectures where the system call is direct,
+        // and there's no socketcall(2) multiplexer indirection.
         #[allow(clippy::cast_possible_truncation)]
         #[allow(clippy::cast_sign_loss)]
         #[allow(clippy::useless_conversion)]
@@ -865,6 +885,116 @@ impl Worker {
             }
         }
 
+        // accept4(2) may be used only with syscall argument cookies.
+        //
+        // We only enforce this on architectures where the system call is direct,
+        // and there's no socketcall(2) multiplexer indirection.
+        #[allow(clippy::cast_possible_truncation)]
+        #[allow(clippy::cast_sign_loss)]
+        #[allow(clippy::useless_conversion)]
+        #[allow(deprecated)]
+        if let Some(syscall) = SYS_ACCEPT4.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
+            // Secure using syscall argument cookies.
+            let mut rules = vec![];
+            if !flags.allow_unsafe_nocookie() {
+                rules.extend(&[
+                    scmp_cmp!($arg4 == (*ACCEPT4_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*ACCEPT4_COOKIE_ARG5).into()),
+                ]);
+            }
+
+            if rules.is_empty() {
+                ctx.add_rule(ScmpAction::Allow, syscall)?;
+            } else {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
+            }
+        } else {
+            match ScmpSyscall::from_name("accept4") {
+                Ok(syscall) => {
+                    // Allow socketcall(2).
+                    ctx.add_rule(ScmpAction::Allow, syscall)?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_emu_syscall",
+                        "msg": "invalid or unsupported syscall accept4");
+                }
+            }
+        }
+
+        // bind(2) may be used only with syscall argument cookies.
+        //
+        // We only enforce this on architectures where the system call is direct,
+        // and there's no socketcall(2) multiplexer indirection.
+        #[allow(clippy::cast_possible_truncation)]
+        #[allow(clippy::cast_sign_loss)]
+        #[allow(clippy::useless_conversion)]
+        #[allow(deprecated)]
+        if let Some(syscall) = SYS_BIND.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
+            // Secure using syscall argument cookies.
+            let mut rules = vec![];
+            if !flags.allow_unsafe_nocookie() {
+                rules.extend(&[
+                    scmp_cmp!($arg3 == (*BIND_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*BIND_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*BIND_COOKIE_ARG5).into()),
+                ]);
+            }
+
+            if rules.is_empty() {
+                ctx.add_rule(ScmpAction::Allow, syscall)?;
+            } else {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
+            }
+        } else {
+            match ScmpSyscall::from_name("bind") {
+                Ok(syscall) => {
+                    // Allow socketcall(2).
+                    ctx.add_rule(ScmpAction::Allow, syscall)?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_emu_syscall",
+                        "msg": "invalid or unsupported syscall bind");
+                }
+            }
+        }
+
+        // connect(2) may be used only with syscall argument cookies.
+        //
+        // We only enforce this on architectures where the system call is direct,
+        // and there's no socketcall(2) multiplexer indirection.
+        #[allow(clippy::cast_possible_truncation)]
+        #[allow(clippy::cast_sign_loss)]
+        #[allow(clippy::useless_conversion)]
+        #[allow(deprecated)]
+        if let Some(syscall) = SYS_CONNECT.map(|n| ScmpSyscall::from_raw_syscall(n as RawSyscall)) {
+            // Secure using syscall argument cookies.
+            let mut rules = vec![];
+            if !flags.allow_unsafe_nocookie() {
+                rules.extend(&[
+                    scmp_cmp!($arg3 == (*CONNECT_COOKIE_ARG3).into()),
+                    scmp_cmp!($arg4 == (*CONNECT_COOKIE_ARG4).into()),
+                    scmp_cmp!($arg5 == (*CONNECT_COOKIE_ARG5).into()),
+                ]);
+            }
+
+            if rules.is_empty() {
+                ctx.add_rule(ScmpAction::Allow, syscall)?;
+            } else {
+                ctx.add_rule_conditional(ScmpAction::Allow, syscall, &rules)?;
+            }
+        } else {
+            match ScmpSyscall::from_name("connect") {
+                Ok(syscall) => {
+                    // Allow socketcall(2).
+                    ctx.add_rule(ScmpAction::Allow, syscall)?;
+                }
+                Err(_) => {
+                    info!("ctx": "confine", "op": "allow_emu_syscall",
+                        "msg": "invalid or unsupported syscall connect");
+                }
+            }
+        }
+
         // Allow futex system calls.
         for sysname in FUTEX_SYSCALLS {
             match ScmpSyscall::from_name(sysname) {
